def declare_variables(variables, macro):
    """
    This is the hook for the functions

    - variables: the dictionary that contains the variables
    - macro: a decorator function, to declare a macro.
    """

    @macro
    def inputcode(filename, language, startline=0, endline=None):
        filename = '../' + filename  # file path must be given relative to root directory
        f = open(filename, 'r')
        if startline != 0 or endline != None:
            lines = f.readlines()
            lines = lines[startline:endline]
            text = "".join(lines)
        else:
            text = f.read()
        textblock = f'```{language}\n{text}\n```'
        return textblock

    @macro
    def inputcpp(filename, startline=0, endline=None):
        return inputcode(filename, 'cpp', startline, endline)



import os, warnings, functools

# Bring all entities from the extension module into this namespace
from .teqp import *
from .teqp import _make_model, _build_multifluid_mutant, _build_multifluid_ecs_mutant

def get_datapath():
    """Get the absolute path to the folder containing the root of multi-fluid data"""
    return os.path.abspath(os.path.dirname(__file__)+"/fluiddata")

from .teqp import __version__

deprecated_top_level_functions = ["get_splus","get_pr","get_B2vir""get_B12vir","pure_VLE_T","extrapolate_from_critical","build_Psir_Hessian_autodiff","build_Psi_Hessian_autodiff","build_Psir_gradient_autodiff","build_d2PsirdTdrhoi_autodiff","get_chempotVLE_autodiff","get_dchempotdT_autodiff","get_fugacity_coefficients","get_partial_molar_volumes","trace_critical_arclength_binary","get_criticality_conditions","eigen_problem","get_minimum_eigenvalue_Psi_Hessian","get_drhovec_dT_crit","get_pure_critical_conditions_Jacobian","solve_pure_critical","mix_VLE_Tx","mixture_VLE_px","get_drhovecdp_Tsat","trace_VLE_isotherm_binary","get_drhovecdT_psat","trace_VLE_isobar_binary","get_dpsat_dTsat_isopleth","mix_VLLE_T","find_VLLE_T_binary"]

def deprecated_caller(model, *args, **kwargs):
    name = kwargs.pop('name123456')
    warnings.warn("Calling the top-level function " + name + " is deprecated and much slower than calling the same-named method of the model instance", FutureWarning)
    return getattr(model, name)(*args, **kwargs)
    
for f in deprecated_top_level_functions:
    globals()[f] = functools.partial(deprecated_caller, name123456=f)
    
# for factory_function in ['_make_vdW1']:

def tolist(a):
    if isinstance(a, list):
        return a
    else:
        try:
            return a.tolist()
        except:
            return a
        
def make_vdW1(a, b):
    AS = _make_model({"kind": "vdW1", "model": {"a": a, "b": b}})
    attach_model_specific_methods(AS)
    return AS
    
def vdWEOS1(*args):
    return make_vdW1(*args)
    
def make_model(*args, **kwargs):
    """
    This function is in two parts; first the make_model function (renamed to _make_model in the Python interface)
    is used to make the model and then the model-specific methods are attached to the instance
    """
    AS = _make_model(*args, **kwargs)
    attach_model_specific_methods(AS)
    return AS

def vdWEOS(Tc_K, pc_Pa):
    j = {
        "kind": "vdW",
        "model": {
            "Tcrit / K": tolist(Tc_K),
            "pcrit / Pa": tolist(pc_Pa)
        }
    }
    return make_model(j)
    
def canonical_PR(Tc_K, pc_Pa, acentric, kmat=None):
    j = {
        "kind": "PR",
        "model": {
            "Tcrit / K": tolist(Tc_K),
            "pcrit / Pa": tolist(pc_Pa),
            "acentric": tolist(acentric),
            "kmat": tolist(kmat) if kmat is not None else None
        }
    }
    return make_model(j)
    
def canonical_SRK(Tc_K, pc_Pa, acentric, kmat=None):
    j = {
        "kind": "SRK",
        "model": {
            "Tcrit / K": tolist(Tc_K),
            "pcrit / Pa": tolist(pc_Pa),
            "acentric": tolist(acentric),
            "kmat": tolist(kmat) if kmat is not None else None
        }
    }
    return make_model(j)

def CPAfactory(spec):
    j = {
        "kind": "CPA",
        "model": spec
    }
    return make_model(j)
    
def PCSAFTEOS(coeffs, kmat=None):
    if isinstance(coeffs[0], SAFTCoeffs):
        coeffs_ = []
        for c in coeffs:
            coeffs_.append({
                'name': c.name,
                'm': c.m,
                'sigma_Angstrom': c.sigma_Angstrom,
                'epsilon_over_k': c.epsilon_over_k,
                'BibTeXKey': c.BibTeXKey
            })
        spec = {'coeffs': coeffs_, 'kmat': tolist(kmat) if kmat is not None else None}
    else:
        spec = {'names': coeffs, 'kmat': tolist(kmat) if kmat is not None else None}
    
    j = {
        "kind": "PCSAFT",
        "model": spec
    }
    return make_model(j)
    
def AmmoniaWaterTillnerRoth():
    return make_model({"kind": "AmmoniaWaterTillnerRoth", "model": {}})

def build_LJ126_TholJPCRD2016():
    return make_model({"kind": "LJ126_TholJPCRD2016", "model": {}})

def IdealHelmholtz(model):
    return make_model({"kind": "IdealHelmholtz", "model": model})
    
def build_multifluid_model(components, coolprop_root, BIPcollectionpath = "", flags = {}, departurepath=""):
    j = {
        "kind": "multifluid",
        "model": {
            "components": components,
            "root": coolprop_root,
            "BIP": BIPcollectionpath,
            "flags": flags,
            "departure": departurepath
        }
    }
    return make_model(j)

def build_multifluid_mutant(*args, **kwargs):
    AS = _build_multifluid_mutant(*args, **kwargs)
    attach_model_specific_methods(AS)
    return AS

def build_multifluid_ecs_mutant(*args, **kwargs):
    AS = _build_multifluid_ecs_mutant(*args, **kwargs)
    attach_model_specific_methods(AS)
    return AS



#!/usr/bin/env python3

#              Copyright Catch2 Authors
# Distributed under the Boost Software License, Version 1.0.
#   (See accompanying file LICENSE_1_0.txt or copy at
#        https://www.boost.org/LICENSE_1_0.txt)

# SPDX-License-Identifier: BSL-1.0

from ConfigureTestsCommon import configure_and_build, run_and_return_output

import os
import re
import sys

"""
Tests the CMake configure option for CATCH_CONFIG_EXPERIMENTAL_REDIRECT

Requires 2 arguments, path folder where the Catch2's main CMakeLists.txt
exists, and path to where the output files should be stored.
"""

if len(sys.argv) != 3:
    print('Wrong number of arguments: {}'.format(len(sys.argv)))
    print('Usage: {} catch2-top-level-dir base-build-output-dir'.format(sys.argv[0]))
    exit(1)

catch2_source_path = os.path.abspath(sys.argv[1])
build_dir_path = os.path.join(os.path.abspath(sys.argv[2]), 'CMakeConfigTests', 'ExperimentalRedirect')

configure_and_build(catch2_source_path,
                    build_dir_path,
                    [("CATCH_CONFIG_EXPERIMENTAL_REDIRECT", "ON")])

stdout, _ = run_and_return_output(os.path.join(build_dir_path, 'tests'),
                                  'SelfTest',
                                  ['-r', 'xml', '"has printf"'])


# The print from printf must be within the XML's reporter stdout tag.
required_output = '''\
      <StdOut>
loose text artifact
      </StdOut>
'''
if not required_output in stdout:
    print("Could not find '{}' in the stdout".format(required_output))
    print('stdout: "{}"'.format(stdout))
    exit(2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
from cpt.packager import ConanMultiPackager
from cpt.ci_manager import CIManager
from cpt.printer import Printer


class BuilderSettings(object):
    @property
    def username(self):
        """ Set catchorg as package's owner
        """
        return os.getenv("CONAN_USERNAME", "catchorg")

    @property
    def login_username(self):
        """ Set Bintray login username
        """
        return os.getenv("CONAN_LOGIN_USERNAME", "horenmar")

    @property
    def upload(self):
        """ Set Catch2 repository to be used on upload.
            The upload server address could be customized by env var
            CONAN_UPLOAD. If not defined, the method will check the branch name.
            Only devel or CONAN_STABLE_BRANCH_PATTERN will be accepted.
            The devel branch will be pushed to testing channel, because it does
            not match the stable pattern. Otherwise it will upload to stable
            channel.
        """
        return os.getenv("CONAN_UPLOAD", "https://api.bintray.com/conan/catchorg/catch2")

    @property
    def upload_only_when_stable(self):
        """ Force to upload when running over tag branch
        """
        return os.getenv("CONAN_UPLOAD_ONLY_WHEN_STABLE", "True").lower() in ["true", "1", "yes"]

    @property
    def stable_branch_pattern(self):
        """ Only upload the package the branch name is like a tag
        """
        return os.getenv("CONAN_STABLE_BRANCH_PATTERN", r"v\d+\.\d+\.\d+")

    @property
    def reference(self):
        """ Read project version from branch create Conan reference
        """
        return os.getenv("CONAN_REFERENCE", "catch2/{}".format(self._version))

    @property
    def channel(self):
        """ Default Conan package channel when not stable
        """
        return os.getenv("CONAN_CHANNEL", "testing")

    @property
    def _version(self):
        """ Get version name from cmake file
        """
        pattern = re.compile(r"project\(Catch2 LANGUAGES CXX VERSION (\d+\.\d+\.\d+)\)")
        version = "latest"
        with open("CMakeLists.txt") as file:
            for line in file:
                result = pattern.search(line)
                if result:
                    version = result.group(1)
        return version

    @property
    def _branch(self):
        """ Get branch name from CI manager
        """
        printer = Printer(None)
        ci_manager = CIManager(printer)
        return ci_manager.get_branch()


if __name__ == "__main__":
    settings = BuilderSettings()
    builder = ConanMultiPackager(
        reference=settings.reference,
        channel=settings.channel,
        upload=settings.upload,
        upload_only_when_stable=False,
        stable_branch_pattern=settings.stable_branch_pattern,
        login_username=settings.login_username,
        username=settings.username,
        test_folder=os.path.join(".conan", "test_package"))
    builder.add()
    builder.run()



# -*- coding: utf-8 -*-
# This file is part of Eigen, a lightweight C++ template library
# for linear algebra.
#
# Copyright (C) 2009 Benjamin Schindler <bschindler@inf.ethz.ch>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Pretty printers for Eigen::Matrix
# This is still pretty basic as the python extension to gdb is still pretty basic. 
# It cannot handle complex eigen types and it doesn't support many of the other eigen types
# This code supports fixed size as well as dynamic size matrices

# To use it:
#
# * Create a directory and put the file as well as an empty __init__.py in 
#   that directory.
# * Create a ~/.gdbinit file, that contains the following:
#      python
#      import sys
#      sys.path.insert(0, '/path/to/eigen/printer/directory')
#      from printers import register_eigen_printers
#      register_eigen_printers (None)
#      end

import gdb
import re
import itertools
from bisect import bisect_left

# Basic row/column iteration code for use with Sparse and Dense matrices
class _MatrixEntryIterator(object):
	
	def __init__ (self, rows, cols, rowMajor):
		self.rows = rows
		self.cols = cols
		self.currentRow = 0
		self.currentCol = 0
		self.rowMajor = rowMajor

	def __iter__ (self):
		return self

	def next(self):
                return self.__next__()  # Python 2.x compatibility

	def __next__(self):
		row = self.currentRow
		col = self.currentCol
		if self.rowMajor == 0:
			if self.currentCol >= self.cols:
				raise StopIteration
				
			self.currentRow = self.currentRow + 1
			if self.currentRow >= self.rows:
				self.currentRow = 0
				self.currentCol = self.currentCol + 1
		else:
			if self.currentRow >= self.rows:
				raise StopIteration
				
			self.currentCol = self.currentCol + 1
			if self.currentCol >= self.cols:
				self.currentCol = 0
				self.currentRow = self.currentRow + 1

		return (row, col)

class EigenMatrixPrinter:
	"Print Eigen Matrix or Array of some kind"

	def __init__(self, variety, val):
		"Extract all the necessary information"
		
		# Save the variety (presumably "Matrix" or "Array") for later usage
		self.variety = variety
		
		# The gdb extension does not support value template arguments - need to extract them by hand
		type = val.type
		if type.code == gdb.TYPE_CODE_REF:
			type = type.target()
		self.type = type.unqualified().strip_typedefs()
		tag = self.type.tag
		regex = re.compile('\<.*\>')
		m = regex.findall(tag)[0][1:-1]
		template_params = m.split(',')
		template_params = [x.replace(" ", "") for x in template_params]
		
		if template_params[1] == '-0x00000000000000001' or template_params[1] == '-0x000000001' or template_params[1] == '-1':
			self.rows = val['m_storage']['m_rows']
		else:
			self.rows = int(template_params[1])
		
		if template_params[2] == '-0x00000000000000001' or template_params[2] == '-0x000000001' or template_params[2] == '-1':
			self.cols = val['m_storage']['m_cols']
		else:
			self.cols = int(template_params[2])
		
		self.options = 0 # default value
		if len(template_params) > 3:
			self.options = template_params[3];
		
		self.rowMajor = (int(self.options) & 0x1)
		
		self.innerType = self.type.template_argument(0)
		
		self.val = val
		
		# Fixed size matrices have a struct as their storage, so we need to walk through this
		self.data = self.val['m_storage']['m_data']
		if self.data.type.code == gdb.TYPE_CODE_STRUCT:
			self.data = self.data['array']
			self.data = self.data.cast(self.innerType.pointer())
			
	class _iterator(_MatrixEntryIterator):
		def __init__ (self, rows, cols, dataPtr, rowMajor):
			super(EigenMatrixPrinter._iterator, self).__init__(rows, cols, rowMajor)

			self.dataPtr = dataPtr

		def __next__(self):
			
			row, col = super(EigenMatrixPrinter._iterator, self).__next__()
			
			item = self.dataPtr.dereference()
			self.dataPtr = self.dataPtr + 1
			if (self.cols == 1): #if it's a column vector
				return ('[%d]' % (row,), item)
			elif (self.rows == 1): #if it's a row vector
				return ('[%d]' % (col,), item)
			return ('[%d,%d]' % (row, col), item)
			
	def children(self):
		
		return self._iterator(self.rows, self.cols, self.data, self.rowMajor)
		
	def to_string(self):
		return "Eigen::%s<%s,%d,%d,%s> (data ptr: %s)" % (self.variety, self.innerType, self.rows, self.cols, "RowMajor" if self.rowMajor else  "ColMajor", self.data)

class EigenSparseMatrixPrinter:
	"Print an Eigen SparseMatrix"

	def __init__(self, val):
		"Extract all the necessary information"

		type = val.type
		if type.code == gdb.TYPE_CODE_REF:
			type = type.target()
		self.type = type.unqualified().strip_typedefs()
		tag = self.type.tag
		regex = re.compile('\<.*\>')
		m = regex.findall(tag)[0][1:-1]
		template_params = m.split(',')
		template_params = [x.replace(" ", "") for x in template_params]

		self.options = 0
		if len(template_params) > 1:
			self.options = template_params[1];
		
		self.rowMajor = (int(self.options) & 0x1)
		
		self.innerType = self.type.template_argument(0)
		
		self.val = val

		self.data = self.val['m_data']
		self.data = self.data.cast(self.innerType.pointer())

	class _iterator(_MatrixEntryIterator):
		def __init__ (self, rows, cols, val, rowMajor):
			super(EigenSparseMatrixPrinter._iterator, self).__init__(rows, cols, rowMajor)

			self.val = val
			
		def __next__(self):
			
			row, col = super(EigenSparseMatrixPrinter._iterator, self).__next__()
				
			# repeat calculations from SparseMatrix.h:
			outer = row if self.rowMajor else col
			inner = col if self.rowMajor else row
			start = self.val['m_outerIndex'][outer]
			end = ((start + self.val['m_innerNonZeros'][outer]) if self.val['m_innerNonZeros'] else
			       self.val['m_outerIndex'][outer+1])

			# and from CompressedStorage.h:
			data = self.val['m_data']
			if start >= end:
				item = 0
			elif (end > start) and (inner == data['m_indices'][end-1]):
				item = data['m_values'][end-1]
			else:
				# create Python index list from the target range within m_indices
				indices = [data['m_indices'][x] for x in range(int(start), int(end)-1)]
				# find the index with binary search
				idx = int(start) + bisect_left(indices, inner)
				if ((idx < end) and (data['m_indices'][idx] == inner)):
					item = data['m_values'][idx]
				else:
					item = 0

			return ('[%d,%d]' % (row, col), item)

	def children(self):
		if self.data:
			return self._iterator(self.rows(), self.cols(), self.val, self.rowMajor)

		return iter([])   # empty matrix, for now


	def rows(self):
		return self.val['m_outerSize'] if self.rowMajor else self.val['m_innerSize']

	def cols(self):
		return self.val['m_innerSize'] if self.rowMajor else self.val['m_outerSize']

	def to_string(self):

		if self.data:
			status = ("not compressed" if self.val['m_innerNonZeros'] else "compressed")
		else:
			status = "empty"
		dimensions  = "%d x %d" % (self.rows(), self.cols())
		layout      = "row" if self.rowMajor else "column"

		return "Eigen::SparseMatrix<%s>, %s, %s major, %s" % (
			self.innerType, dimensions, layout, status )

class EigenQuaternionPrinter:
	"Print an Eigen Quaternion"
	
	def __init__(self, val):
		"Extract all the necessary information"
		# The gdb extension does not support value template arguments - need to extract them by hand
		type = val.type
		if type.code == gdb.TYPE_CODE_REF:
			type = type.target()
		self.type = type.unqualified().strip_typedefs()
		self.innerType = self.type.template_argument(0)
		self.val = val
		
		# Quaternions have a struct as their storage, so we need to walk through this
		self.data = self.val['m_coeffs']['m_storage']['m_data']['array']
		self.data = self.data.cast(self.innerType.pointer())
			
	class _iterator:
		def __init__ (self, dataPtr):
			self.dataPtr = dataPtr
			self.currentElement = 0
			self.elementNames = ['x', 'y', 'z', 'w']
			
		def __iter__ (self):
			return self
	
		def next(self):
			return self.__next__()  # Python 2.x compatibility

		def __next__(self):
			element = self.currentElement
			
			if self.currentElement >= 4: #there are 4 elements in a quanternion
				raise StopIteration
			
			self.currentElement = self.currentElement + 1
			
			item = self.dataPtr.dereference()
			self.dataPtr = self.dataPtr + 1
			return ('[%s]' % (self.elementNames[element],), item)
			
	def children(self):
		
		return self._iterator(self.data)
	
	def to_string(self):
		return "Eigen::Quaternion<%s> (data ptr: %s)" % (self.innerType, self.data)

def build_eigen_dictionary ():
	pretty_printers_dict[re.compile('^Eigen::Quaternion<.*>$')] = lambda val: EigenQuaternionPrinter(val)
	pretty_printers_dict[re.compile('^Eigen::Matrix<.*>$')] = lambda val: EigenMatrixPrinter("Matrix", val)
	pretty_printers_dict[re.compile('^Eigen::SparseMatrix<.*>$')] = lambda val: EigenSparseMatrixPrinter(val)
	pretty_printers_dict[re.compile('^Eigen::Array<.*>$')]  = lambda val: EigenMatrixPrinter("Array",  val)

def register_eigen_printers(obj):
	"Register eigen pretty-printers with objfile Obj"

	if obj == None:
		obj = gdb
	obj.pretty_printers.append(lookup_function)

def lookup_function(val):
	"Look-up and return a pretty-printer that can print va."
	
	type = val.type
	
	if type.code == gdb.TYPE_CODE_REF:
		type = type.target()
	
	type = type.unqualified().strip_typedefs()
	
	typename = type.tag
	if typename == None:
		return None
	
	for function in pretty_printers_dict:
		if function.search(typename):
			return pretty_printers_dict[function](val)
	
	return None

pretty_printers_dict = {}

build_eigen_dictionary ()



#!/usr/bin/env python
from conan import ConanFile, tools, __version__ as conan_version
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps, cmake_layout
from conan.tools import files, scm
import os
import shutil
import re

required_conan_version = ">=1.53.0"

class CatchConan(ConanFile):
    name = "catch2"
    description = "A modern, C++-native, framework for unit-tests, TDD and BDD"
    topics = ("conan", "catch2", "unit-test", "tdd", "bdd")
    url = "https://github.com/catchorg/Catch2"
    homepage = url
    license = "BSL-1.0"
    version = "latest"

    exports = "LICENSE.txt"
    exports_sources = ("src/*", "CMakeLists.txt", "CMake/*", "extras/*")

    settings = "os", "compiler", "build_type", "arch"

    def set_version(self):
        pattern = re.compile(r"\w*VERSION (\d+\.\d+\.\d+) # CML version placeholder, don't delete")
        with open("CMakeLists.txt") as file:
            for line in file:
                result = pattern.search(line)
                if result:
                    self.version = result.group(1)

        self.output.info(f'Using version: {self.version}')

    def layout(self):
        cmake_layout(self)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()

        deps = CMakeDeps(self)
        deps.generate()

    def _configure_cmake(self):
        cmake = CMake(self)

        # These are option variables. The toolchain in conan 2 doesn't appear to
        # set these correctly so you have to do it in the configure variables.
        cmake.configure(variables= {
            "BUILD_TESTING": "OFF",
            "CATCH_INSTALL_DOCS": "OFF",
            "CATCH_INSTALL_EXTRAS": "ON",
            }
        )
        return cmake

    def build(self):
        cmake = self._configure_cmake()
        cmake.build()

    def package(self):
        cmake = self._configure_cmake()
        cmake.install()

        os.mkdir(f'{self.package_folder}/licenses/')
        shutil.copy2(f'{self.recipe_folder}/LICENSE.txt', f'{self.package_folder}/licenses/')

    def package_info(self):
        lib_suffix = "d" if self.settings.build_type == "Debug" else ""

        self.cpp_info.set_property("cmake_file_name", "Catch2")
        self.cpp_info.set_property("cmake_target_name", "Catch2::Catch2WithMain")
        self.cpp_info.set_property("pkg_config_name", "catch2-with-main")

        # Catch2
        self.cpp_info.components["catch2base"].set_property("cmake_file_name", "Catch2::Catch2")
        self.cpp_info.components["catch2base"].set_property("pkg_config_name", "catch2")
        self.cpp_info.components["catch2base"].libs = ["Catch2" + lib_suffix]
        self.cpp_info.components["catch2base"].builddirs.append("lib/cmake/Catch2")

        # Catch2WithMain
        self.cpp_info.components["catch2main"].set_property("cmake_file_name", "Catch2::Catch2WithMain")
        self.cpp_info.components["catch2main"].set_property("cmake_target_name", "Catch2::Catch2WithMain")
        self.cpp_info.components["catch2main"].set_property("pkg_config_name", "catch2-with-main")
        self.cpp_info.components["catch2main"].libs = ["Catch2Main" + lib_suffix]
        self.cpp_info.components["catch2main"].requires = ["catch2base"]


#!/usr/bin/env python3

from scriptCommon import catchPath
import os
import subprocess

# ---------------------------------------------------
# Update code examples
# ---------------------------------------------------
# For info on mdsnippets, see https://github.com/SimonCropp/MarkdownSnippets

# install dotnet SDK from http://go.microsoft.com/fwlink/?LinkID=798306&clcid=0x409
# Then install MarkdownSnippets.Tool with
#   dotnet tool install -g MarkdownSnippets.Tool
# To update:
#   dotnet tool update  -g MarkdownSnippets.Tool
# To uninstall (e.g. to downgrade to a lower version)
# dotnet tool uninstall -g MarkdownSnippets.Tool

os.chdir(catchPath)

subprocess.run('dotnet tool update  -g MarkdownSnippets.Tool --version 21.2.0', shell=True, check=True)
subprocess.run('mdsnippets', shell=True, check=True)



import numpy as np
import numbers

class MultiComplex:
    """

    Special methods: https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
    """
    def __init__(self, *, a=None, b=None, d=1, v=None):

        if v is not None:
            L = len(v)
            if L < 1:
                raise ValueError("Doesn't make sense")
            elif L == 1:
                self.a = v[0]
                self.b = 0
                self.d = 1
            elif L == 2:
                self.a = v[0]
                self.b = v[1]
                self.d = 1
            elif L % 2 != 0:
                raise ValueError("Length of coeffs [{0:d}] must be power of 2".format(L))
            else:
                r = MultiComplex(v=v[0:L//2])
                s = MultiComplex(v=v[L//2::])
                self.a = r
                self.b = s
                self.d = r.d+1
        else:
            self.a = a
            self.b = b
            self.d = d

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        """
        Overload some methods from numpy to handle multicomplex numbers

        N.B.: Not all numpy methods are implemented as ufunc (e.g., np.real), and therefore, 
        they will just return the object back again.

        The expansion for sin goes something like this in sympy:
        >>> a, b = symbols('a,b', real=True)
        >>> sin(a + I*b).expand(complex=True)
        """
        if method == '__call__':
            z = self
            if ufunc is np.cos:
                a, b = reim(z); return MultiComplex(a=np.cos(a)*np.cosh(b),b=-np.sin(a)*np.sinh(b), d=dim(z))
            elif ufunc is np.sin:
                a, b = reim(z); return MultiComplex(a=np.sin(a)*np.cosh(b), b=np.cos(a)*np.sinh(b), d=dim(z))
            elif ufunc is np.sinh:
                a, b = reim(z); return MultiComplex(a=np.sinh(a)*np.cos(b), b=np.cosh(a)*np.sin(b), d=dim(z))
            elif ufunc is np.cosh:
                a, b = reim(z); return MultiComplex(a=np.cosh(a)*np.cos(b), b=np.sinh(a)*np.sin(b), d=dim(z))
            elif ufunc is np.exp:
                a, b = reim(z); return np.exp(a) * MultiComplex(a=np.cos(b), b=np.sin(b), d=dim(z))

    def __neg__(self):
        z = self
        return MultiComplex(a=-real(z), b=-imag(z), d=dim(z))

    def __mul__(self, w):
        z = self
        if dim(z) < dim(w):
            return MultiComplex(a=z*real(w), b=z*imag(w), d=dim(w)) 
        elif dim(z) > dim(w):
            return MultiComplex(a=real(z)*w, b=imag(z)*w, d=dim(z))
        else:
            return MultiComplex(a=real(z)*real(w)-imag(z)*imag(w),
                                b=real(z)*imag(w)+imag(z)*real(w), d=dim(z))

    def __sub__(self, w):
        z = self
        if dim(z) < dim(w):
            return MultiComplex(a=z-real(w),b=-imag(w),d=dim(w)) 
        elif dim(z) > dim(w):
            return MultiComplex(a=real(z)-w,b=imag(z),d=dim(z))
        else:
            return MultiComplex(a=real(z)-real(w),b=imag(z)-imag(w),d=dim(z))

    def __add__(self, w):
        z = self
        if isinstance(w, numbers.Number):
            return MultiComplex(a=w+real(z),b=w+imag(z),d=dim(z))
        elif dim(z) < dim(w):
            return MultiComplex(a=z+real(w),b=imag(w),d=dim(w))
        elif dim(z) > dim(w):
            return MultiComplex(a=real(z)+w,b=imag(z),d=dim(z))
        else:
            return MultiComplex(a=real(z)+real(w),b=imag(z)+imag(w),d=dim(z))

    def __getitem__(self, key=None):
        z = self
        if isinstance(key, tuple):
            if dim(z) == 1 and len(key) == 1:
                if key[0] == 0:
                    return real(z)
                elif key[0] == 1:
                    return imag(z)
            elif len(key) > 1 and key[0] == 0:
                return real(z)[key[1::]]
            elif len(key) > 1 and key[0] == 1:
                return imag(z)[key[1::]]
            else:
                raise IndexError("MultiComplex key length of {0:d} must match dim of {1:d}".format(len(key), dim(z)))
        else:
            raise TypeError("Input must be tuple")

    def __repr__(self):
        return str((self.a, self.b, self.d))

def real(z):
    return z.a
def imag(z):
    return z.b
def reim(z):
    return (z.a, z.b)
def dim(z):
    return z.d

rx = 0.1234
m1 = MultiComplex(v=[rx, 1e-100,0,0])
def f(x):
    return np.sin(x)
def dfdx(x):
    return np.cos(x)
m1s = f(m1)
ind_mocom = tuple([0,1]) # most complex index
print(dfdx(rx), m1s[ind_mocom]/1e-100)
print(m1s[(0,0)])
print(m1s[(0,1)])
print(m1s[(1,0)])
print(m1s[(1,1)])

rx = 0.1
m1 = MultiComplex(v=[rx, 1e-100]+[0]*14)

def f(x):
    return x*np.sin(x)

def dfdx(x):
    return np.sin(x) + x*np.cos(x)

m1s = f(m1)
ind_mocom = tuple([0]*(m1.d-1)+[1]) # most complex index
print(dfdx(rx), m1s[ind_mocom]/1e-100)


import codecs
import re
import json
import glob
import os
import math
import hashlib

import numpy as np
import CoolProp.CoolProp as CP

def lines_starts_with(s, lines, *, start_index=0):
    matches = []
    for i in range(start_index, len(lines)):
        if lines[i].startswith(s):
            matches.append(i)
    return matches

def lines_contains(s, lines, *, start_index=0):
    matches = []
    for i in range(start_index, len(lines)):
        if s in lines[i]:
            matches.append(i)
    return matches

def lines_equals(s, lines, *, start_index=0):
    matches = []
    for i in range(start_index, len(lines)):
        if re.fullmatch(s, lines[i]):
            matches.append(i)
    return matches


class FLDDeconstructor:
    """
    Read a REFPROP fluid file, and convert to the 
    JSON format of CoolProp
    """
    def __init__(self, path, encoding='cp1252'):
        """
        Load in the FLD file, do some pre-processing
        """
        self.path = path
        with codecs.open(path, encoding=encoding) as fp:
            self.contents = fp.read()
            self.lines = self.contents.split('\n')

    def get_keyed_line(self, block, key, conv, *, alias=None):
        keys = [key]
        if alias:
            keys += [alias]
        for k in keys:
            if not k in ''.join(block): continue
            lines = list(filter(lambda l: k in l, block))
            if len(lines) == 1:
                els = lines[0].split('!')[0]
                return [conv(el) for el in els.split(' ') if el]
        raise KeyError(key)

    
    def get_block(self, startreg, endreg, start_index=0, indices=False):
        matches = lines_starts_with(startreg, self.lines, start_index = start_index)
        istart = matches[0]
        matches = lines_equals(endreg, self.lines, start_index = istart+1)
        iend = matches[0]
        if indices:
            return self.lines[istart:iend], istart, iend
        else:
            return self.lines[istart:iend]

    def strip_commented(self, block):
        return list(filter(lambda l: not l.startswith('?') and not l.startswith('!'), block))

    def strip_comments(self, block):
        block = list(filter(lambda l: not l.startswith('?') and not l.startswith('!'), block))
        for iline, line in enumerate(block):
            if '!' in line:
                block[iline] = line.split('!')[0].rstrip() + '\n'
        return block

    def get_ancillary_description(self, key):
        return {
        # Most are in this form:
        'PS5':  'P=Pc*EXP[SUM(Ni*Theta^ti)*Tc/T]',
        'DL1':  'D=Dc*[1+SUM(Ni*Theta^ti)]',
        'DV3':  'D=Dc*EXP[SUM(Ni*Theta^ti)]',

        # Other ones might be of this form:
        'DL2':  'D=Dc*[1+SUM(Ni*Theta^(ti/3))]',
        'DL4':  'D=Dc*EXP[SUM(Ni*Theta^(ti/3))]',
        'DV6':  'D=Dc*EXP[SUM(Ni*Theta^(ti/3))*Tc/T]',
        'DV4':  'D=Dc*EXP[SUM(Ni*Theta^(ti/3))]',
        'DL6':  'D=Dc*EXP[SUM(Ni*Theta^(ti/3))*Tc/T]',
        }[key]

    def get_ancillary(self, key):
        block = self.get_block('#'+key, r'\s*')
        block = self.strip_comments(block)
        model_key = block[1].strip()
        limits = [float(block[i]) for i in range(2, 6)]
        reducing = [float(v) for v in block[6].strip().split(' ') if v]
        Ncoeffs = [int(v) for v in block[7].split(' ') if v]
        n, t = [], []
        reducing[1] *= 1000 # REFPROP uses mol/L & kPa, CoolProp uses mol/m^3 & Pa, multiply by 1000 to convert
        for row in block[8::]:
            _n, _t = [_ for _ in row.lstrip().split(' ') if _][0:2]
            n.append(float(_n.replace('D+','e+')))
            t.append(float(_t))
        assert(len(n) == Ncoeffs[0] and len(t) == Ncoeffs[0])
        desc = self.get_ancillary_description(model_key.strip())
        if model_key[0:2] == 'PS':
            type_key = 'pL'
        elif model_key in ['DL1','DL2']:
            type_key = 'rhoLnoexp'
        elif model_key[0:2] == 'DL':
            type_key = 'rhoL'
            t = t if model_key[2] in ['1','3','5'] else [t_/3 for t_ in t]
        elif model_key[0:2] == 'DV':
            type_key = 'rhoV'
            t = t if model_key[2] in ['1','3','5'] else [t_/3 for t_ in t]
        else:
            raise KeyError(model_key)

        return {
            "T_r": reducing[0],
            "Tmax": reducing[0],
            "Tmin": limits[0],
            'description': desc,
            'n': n,
            'reducing_value': reducing[1],
            't': t,
            'type': type_key,
            'using_tau_r': 'Tc/T' in desc
        }

    def get_all_ancillaries(self):
        return{
            'PS': self.get_ancillary('PS'),
            'DV': self.get_ancillary('DV'),
            'DL': self.get_ancillary('DL')
        }

    def get_ancillaries(self):
        anc = self.get_all_ancillaries()
        return {'pS': anc['PS'], 'rhoL': anc['DL'], 'rhoV': anc["DV"]}

    def formula_from_inchi(self):
        """ Parse the standard inchi string to extract the chemical formula in Hill order """
        try:
            stdinchistring = self.get_keyed_line(self.lines, '!Standard InChI String', lambda x: x)[0]
        except:
            return 'COULD NOT PARSE InChI STRING'
        if '/' not in stdinchistring:
            raise ValueError(f'{stdinchistring} is not a valid standard InChI key')
        formula = stdinchistring.split('/')[1]
        matches = re.findall(r'[A-Z][a-z]*[0-9]*', formula)
        assert(sum([len(m) for m in matches]) == len(formula)) # make sure all parts are matched
        o = ''
        for match in matches:
            def replacer(m):
                N = len(m.groups())
                # print(m, N)
                if N == 0:
                    return m.group(0)+'_{1}'
                else:
                    i = m.group(1)
                    return f'_{{{i}}}'
            newmatch = re.sub(r'([0-9]+)', replacer, match)
            if '_{' not in newmatch:
                newmatch += '_{1}'
            o += newmatch
        return o

    def get_info(self):
        CHEMSPIDER_ID = -1
        i = lines_contains('!Short name', self.lines)
        short_name = self.lines[i[0]].split('!')[0].strip()
        name = os.path.split(self.path)[1].split('.')[0]

        i = lines_contains('!Full name', self.lines)
        full_name = self.lines[i[0]].split('!')[0].strip()

        name = os.path.split(self.path)[1].split('.')[0]
        aliases = [short_name, short_name.lower(), full_name, full_name.lower(), full_name.replace(' ','').upper(), short_name.replace(' ','').upper()]
        aliases = [alias for alias in aliases if alias != name]
        
        StdInChIstr = 'UNKNOWN'
        try:
            StdInChIstr = self.get_keyed_line(self.lines, '!Standard InChI String', lambda x: x)[0]
            if not StdInChIstr.startswith('InChI='):
                StdInChIstr = "InChI=" + StdInChIstr
        except KeyError:
            print('Unable to find InChI key')

        InChIKey = 'UNKNOWN'
        try:
            InChIKey = self.get_keyed_line(self.lines, '!Standard InChI Key', lambda x: x)[0]
        except KeyError:
            print('Unable to find InChI key')

        GWP100 = 'UNKNOWN'
        try:
            GWP100 = float(self.get_keyed_line(self.lines, '!GWP', lambda x: x)[0])
        except KeyError:
            print('Unable to find GWP')

        ODP = 'UNKNOWN'
        try:
            ODP = float(self.get_keyed_line(self.lines, '!ODP', lambda x: x)[0])
        except KeyError:
            print('Unable to find ODP')

        ASHRAE34 = 'UNKNOWN'
        try:
            ASHRAE34 = self.get_keyed_line(self.lines, '!Safety Group', lambda x: x)[0]
        except KeyError:
            print('Unable to find Safety group')
        
        return {
            "2DPNG_URL": f"http://www.chemspider.com/ImagesHandler.ashx?id={CHEMSPIDER_ID}",
            "ALIASES": list(set(aliases)),
            "CAS": self.get_keyed_line(self.lines, '!CAS number', lambda x: x)[0],
            "CHEMSPIDER_ID": CHEMSPIDER_ID,
            "ENVIRONMENTAL": {
              "ASHRAE34": ASHRAE34,
              "FH": 0,
              "GWP100": GWP100,
              "GWP20": 0.0,
              "GWP500": 0.0,
              "HH": 0,
              "Name": name,
              "ODP": ODP,
              "PH": 1e30
            },
            "FORMULA": self.formula_from_inchi(),
            "INCHI_KEY": InChIKey,
            "INCHI_STRING": StdInChIstr,
            "NAME": name,
            "REFPROP_NAME": name,
            "SMILES": "?"
        }

    def get_critical_state(self):

        i = lines_contains('!Critical temperature [K]', self.lines)
        Tc = float(self.lines[i[0]].split('!')[0].strip())

        i = lines_contains('!Critical pressure [kPa]', self.lines)
        pc_kPa = float(self.lines[i[0]].split('!')[0].strip())

        i = lines_contains('!Critical density [mol/L]', self.lines)
        rhoc_moldm3 = float(self.lines[i[0]].split('!')[0].strip())
        
        return {
              "T": Tc,
              "T_units": "K",
              "hmolar": -99999999999,
              "hmolar_units": "J/mol",
              "p": pc_kPa*1e3,
              "p_units": "Pa",
              "rhomolar": rhoc_moldm3*1e3,
              "rhomolar_units": "mol/m^3",
              "smolar": 999999999999999,
              "smolar_units": "J/mol/K"
        }

    def ancillary_evaluator(self, anc, T):
        T_r = anc['T_r']
        theta = 1-T/T_r
        RHS = sum([_n*theta**_t for _n, _t in zip(anc['n'], anc['t'])])
        if anc['using_tau_r']:
            RHS *= T_r/T
        if anc['type'] == 'rhoLnoexp':
            return anc['reducing_value']*(1+RHS)
        elif anc['type'] in ['pV','pL','rhoV','rhoL']:
            return math.exp(RHS)*anc['reducing_value']
        else:
            raise KeyError(anc['type'])

    def check_ancillaries(self, anc, name, T, backend = 'REFPROP'):
        AS = CP.AbstractState(backend, name)
        AS.update(CP.QT_INPUTS, 0, T)
        rhoL = AS.rhomolar()
        rhoV = AS.saturated_vapor_keyed_output(CP.iDmolar)
        p = AS.p()

        try:
            rhoL_anc = self.ancillary_evaluator(anc['DL'], T)
            rhoV_anc = self.ancillary_evaluator(anc['DV'], T)
            p_anc = self.ancillary_evaluator(anc['PS'], T)

            def check_ok(k, err, thresh):
                if err > thresh:
                    raise ValueError(name, k, err, thresh)
            check_ok('DL',abs(rhoL_anc-rhoL)/rhoL, 1e-3)
            check_ok('DV',abs(rhoV_anc-rhoV)/rhoV, 1e-3)
            check_ok('PS',abs(p_anc-p)/p, 1e-3)

        except BaseException as BE:
            print(BE)
            # raise

    def get_alpha0(self, Tc):
        def get_PX0_block():
            imin, imax = 0, len(self.lines)
            for i in range(100):
                try:
                    block, imin, imaxnew = self.get_block('#AUX', r'^\s*[\n\r]', imin, indices=True) # end of block is either empty line or three or more spaces
                    imin = imaxnew
                    kind = block[1].split(' ')[0]
                    if kind == 'PX0':
                        return block
                except IndexError:
                    pass
            raise IndexError("Cannot find PX0 block")
        PX0_block = get_PX0_block()
        # print(''.join(PX0_block))
        term_spec = self.get_keyed_line(PX0_block,' !Nterms',int)
        names = ['ai*log(tau**ti)','ai*tau**ti','ai*log(1-exp(bi*tau))']

        alpha0 = []

        # To get the ln(delta) term that doesn't appear in the block
        # anywhere
        alpha0.append({
            "a1": 0.0,
            "a2": 0.0,
            "type": "IdealGasHelmholtzLead"
        })

        i = lines_contains('!Nterms', PX0_block)[0]+1
        for Nterms, name in zip(term_spec, names):
            lines = PX0_block[i:i+Nterms]
            if name == 'ai*log(tau**ti)':
                if len(lines) == 1:
                    els = [el.strip() for el in lines[0].split(' ') if el]
                    if float(els[1]) != 1.0: 
                        raise ValueError("ti must be 1.0 in ai*log(tau**ti)")
                    alpha0.append({
                        "a": float(els[0]),
                        'type': 'IdealGasHelmholtzLogTau'
                    })
                else:
                    raise IndexError()
            elif name == 'ai*tau**ti':
                # This is a combination of the IdealGasHelmholtzLogTau and IdealGasHelmholtzLogTau terms in CoolProp
                n,t = [],[]
                for line in lines:
                    els = [el.strip() for el in line.split(' ') if el]
                    ai = float(els[0])
                    ti = float(els[1])
                    n.append(ai)
                    t.append(ti)
                alpha0.append({
                    "n": n,
                    "t": t,
                    "type": "IdealGasHelmholtzPower"
                })
            elif name == 'ai*log(1-exp(bi*tau))':
                n, t = [],[]
                for line in lines:
                    els = [el.strip() for el in line.split(' ') if el]
                    ni,ti = [float(el) for el in els[0:2]]
                    n.append(ni)
                    t.append(ti/Tc)
                alpha0.append({
                    "n": n, "t": t,
                    "type": "IdealGasHelmholtzPlanckEinstein"
                })
            elif name == 'Hmm':
                pass
            else:
                raise ValueError("Bad alpha0 contribution" + name)

            i += Nterms

        return alpha0

    def get_EOS(self, ideal=True):
        block = self.get_block('#EOS', r'^\s*[\n\r]') # end of block is a line with only whitespace and terminated with a newline, or a carriage return
        block = self.strip_commented(block)
        indices = lines_contains('eta      beta    gamma   epsilon', block)
        if indices:
            block = block[0:indices[0]]
        # print(''.join(block))
        kind = block[1].split(' ')[0]
        if kind != 'FEQ':
            raise ValueError(f'Kind of EOS [{kind}] is not supported; only valid options are: {{FEQ}}')

        def get_keyed_line(key,conv,*,alias=None):
            return self.get_keyed_line(block, key, conv, alias=alias)

        # Various constants. Anything without a unit suffix is in base SI units
        Tmax = get_keyed_line('!Upper temperature limit [K]',float)[0]
        Tt = get_keyed_line('!Triple point temperature',float)[0]
        acentric = get_keyed_line('!Acentric factor',float)[0]
        R = get_keyed_line('!Gas constant',float,alias='!gas constant')[0]
        molemass_kgkmol = get_keyed_line('!Molar mass',float)[0]
        pt_kPa = get_keyed_line('!Pressure at triple point',float)[0]
        pmax_kPa = get_keyed_line('!Upper pressure limit',float)[0]
        # Reducing parameters
        _Tr, pc_kPa, _rhoc_moldm3 = get_keyed_line('!Tc [K], pc [kPa], rhoc [mol/L]',float)
        Tr, rhor_moldm3 = get_keyed_line('!Reducing parameters [K, mol/L]',float)
        # Term specification
        try:
            terms = get_keyed_line(' !# terms and # coefs/term',int)
            i0 = lines_contains('!# terms and # coefs/term', block)[0]
        except KeyError:
            terms = get_keyed_line(' !First digit: # of normal terms',int)
            i0 = lines_contains(' !First digit: # of normal terms', block)[0]
            
        Npoly, Ncoefpoly, NGaussian, NcoefGaussian, NGao = terms[0:5]
        assert(sum(terms[5::]) == 0)

        pairs = []
        names = []
        if Npoly > 0:
            names.append('Polynomial')
            pairs.append((Npoly, Ncoefpoly))
        if NGaussian > 0:
            names.append('Gaussian')
            pairs.append((NGaussian, NcoefGaussian))
        if NGao > 0:
            names.append('Gao')
            pairs.append((NGao, 12))

        alphar = []
        i = i0 + 1
        for name, pair in zip(names, pairs):

            Nterms, Ncoefperterm = pair
            if Ncoefperterm == 0 and name in ['Polynomial','Gaussian']:
                continue
            lines = block[i:i+Nterms]
            if name == 'Polynomial':
                if Ncoefperterm == 4:
                    n,t,d,l = [],[],[],[]
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        ni,ti,di,li = [float(el) for el in els[0:4]]
                        n.append(ni)
                        t.append(ti)
                        d.append(di)
                        l.append(li)
                    alphar.append({"n": n, "t": t, "d": d, "l": l,
                                   "type": "ResidualHelmholtzPower"})
                elif Ncoefperterm == 5:
                    n,t,d,l,g = [],[],[],[],[]
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        ni,ti,di,li,gi = [float(el) for el in els[0:5]]
                        n.append(ni)
                        t.append(ti)
                        d.append(di)
                        l.append(li)
                        g.append(gi)
                    alphar.append({"n": n, "t": t, "d": d, "l": l, "g": g,
                                   "type": "ResidualHelmholtzExponential"})
                else:
                    raise ValueError()

            elif name == 'Gaussian':

                def is_normal_gaussian(lines):
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        if not (all([float(el)==0 for el in els[9:12]]) and all([float(el)==2.0 for el in els[3:5]])):
                            return False
                    return True

                def is_R125_gaussian(lines):
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        if not (all([float(el)==0 for el in els[9:12]]) and all([float(el)==-1.0 for el in els[5:7]])):
                            return False
                    return True

                def is_nonanalytic_gaussian(lines):
                    numgaussian = 0; numNA = 0
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        if all([float(el)==0 for el in els[9:12]]) and all([float(el)==2.0 for el in els[3:5]]):
                            numgaussian += 1
                        if all([float(el)!=0 for el in els[9:12]]) and all([float(el)==2.0 for el in els[3:5]]):
                            numNA += 1
                    return (numNA > 0, numgaussian, numNA)

                def add_normal_gaussian(lines):
                    n,t,d,eta,beta,gamma,epsilon = [],[],[],[],[],[],[]
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        ni,ti,di,ph1,ph2,negetai,negbetai,gammai,epsiloni,ph3,ph4,ph5 = [float(el) for el in els[0:12]]
                        n.append(ni)
                        t.append(ti)
                        d.append(di)
                        eta.append(-negetai)
                        beta.append(-negbetai)
                        gamma.append(gammai)
                        epsilon.append(epsiloni)
                    alphar.append({"n": n, "t": t, "d": d, 
                               "eta": eta, "beta": beta, "gamma": gamma, "epsilon": epsilon,
                               "type": "ResidualHelmholtzGaussian"})

                def add_nonanalytic(lines):
                    n,a,b,beta,A,B,C,D = [],[],[],[],[],[],[],[]
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        ni,ph1,ph2,ph3,ph4,bi,betai,Ai,Ci,Di,Bi,ai = [float(el) for el in els[0:12]]
                        n.append(ni)
                        a.append(ai)
                        b.append(bi)
                        beta.append(betai)
                        A.append(Ai)
                        B.append(Bi)
                        C.append(Ci)
                        D.append(Di)
                    alphar.append({"n": n, "b": b, "a": a, "b": b, "A": A, "B": B, "C": C, "D": D, "beta": beta,
                               "type": "ResidualHelmholtzNonAnalytic"})

                if is_normal_gaussian(lines):
                    add_normal_gaussian(lines)
                elif is_R125_gaussian(lines):
                    n,t,d,l,m = [],[],[],[],[]
                    for line in lines:
                        els = [el.strip() for el in line.split(' ') if el]
                        # Normal gaussian term if last three placeholders are zero
                        ni,ti,di,li,mi,ph1,ph2,ph3,ph4,ph5,ph6,ph7 = [float(el) for el in els[0:12]]
                        n.append(ni)
                        t.append(ti)
                        d.append(di)
                        l.append(li)
                        m.append(mi)
                    alphar.append({"n": n, "t": t, "d": d, "l": l, "m": m, "type": "ResidualHelmholtzLemmon2005"})
                else:
                    isnonanalytic, NGaussian, NNA = is_nonanalytic_gaussian(lines)
                    if isnonanalytic:
                        add_normal_gaussian(lines[0:NGaussian])
                        add_nonanalytic(lines[NGaussian::])
                    else:
                        raise ValueError("I don't yet understand this kind of Gaussian:"+str(lines))
            elif name == 'Gao':
                n,t,d,eta,beta,gamma,epsilon,b = [],[],[],[],[],[],[],[]
                for line in lines:
                    els = [el.strip() for el in line.split(' ') if el]
                    ni,ti,di,ph1,ph2,etai,betai,gammai,epsiloni,bi,ph4,ph5 = [float(el) for el in els[0:12]]
                    n.append(ni)
                    t.append(ti)
                    d.append(di)
                    eta.append(etai)
                    beta.append(betai)
                    gamma.append(gammai)
                    epsilon.append(epsiloni)
                    b.append(bi)
                alphar.append({"n": n, "t": t, "d": d, 
                           "eta": eta, "beta": beta, "gamma": gamma, "epsilon": epsilon, "b": b,
                           "type": "ResidualHelmholtzGaoB"})
            else:
                raise ValueError("I don't yet understand:"+name)
                
            i += Nterms

        pr = pc_kPa*1e3

        # Get the terms
        EOS = {
          "BibTeX_CP0": "",
          "BibTeX_EOS": "XXX-X-XXXX",
          "STATES": {
            "reducing": {
              "T": Tr,
              "T_units": "K",
              "hmolar": -99999999999,
              "hmolar_units": "J/mol",
              "p": pr,
              "p_units": "Pa",
              "rhomolar": rhor_moldm3*1e3,
              "rhomolar_units": "mol/m^3",
              "smolar": 999999999999999,
              "smolar_units": "J/mol/K"
            },
            "sat_min_liquid": {
              "T": Tt,
              "T_units": "K",
              "hmolar": -999999999999,
              "hmolar_units": "J/mol",
              "p": pt_kPa*1e3,
              "p_units": "Pa",
              "rhomolar": 9999999999999999,
              "rhomolar_units": "mol/m^3",
              "smolar": 99999999999999999999,
              "smolar_units": "J/mol/K"
            },
            "sat_min_vapor": {
              "T": Tt,
              "T_units": "K",
              "hmolar": 9999999999999,
              "hmolar_units": "J/mol",
              "p": pt_kPa*1e3,
              "p_units": "Pa",
              "rhomolar": 99999999999999,
              "rhomolar_units": "mol/m^3",
              "smolar": 9999999999999999999,
              "smolar_units": "J/mol/K"
            }
          },
          "T_max": Tmax,
          "T_max_units": "K",
          "Ttriple": Tt,
          "Ttriple_units": "K",
          "acentric": acentric,
          "acentric_units": "-",
          "alpha0": self.get_alpha0(Tc=Tr) if ideal else [],
          "alphar": alphar,
          "gas_constant": R,
          "gas_constant_units": "J/mol/K",
          "molar_mass": molemass_kgkmol/1e3,
          "molar_mass_units": "kg/mol",
          "p_max": pmax_kPa*1e3,
          "p_max_units": "Pa",
          "pseudo_pure": False
        }
        return EOS

    def write_JSON(self, jsonpath, *, ideal=True, ancillaries=False):
        """ """
        f = {
            'EOS': [self.get_EOS(ideal=ideal)],
            'ANCILLARIES': self.get_ancillaries() if ancillaries else {},
            'INFO': self.get_info(),
        }
        f['STATES'] = {
            'critical': self.get_critical_state(),
            'triple_liquid': f["EOS"][0]['STATES']['sat_min_liquid'],
            'triple_vapor': f["EOS"][0]['STATES']['sat_min_vapor']
        }
        with open(jsonpath, 'w') as fp:
            fp.write(json.dumps(f, indent=2, sort_keys=True))

class HMXDeconstructor:
    """ 
    Take a REFPROP 10 compatible HMX.BNC, convert to 
    CoolProp format departure and binary interaction files
    """
    def __init__(self, path, RProot):
        self.path = path
        self.RProot = RProot
        self.lines = [line.strip() for line in codecs.open(path, encoding='cp1252').readlines()]
        self.CASlookup = self.get_CASlookup()
        self.RPlookup = self.get_RPlookup()
        self.dep = self.get_departure_functions()
        self.BIP = self.get_binary_interaction()

    def get_CASlookup(self):
        o = {}
        for FLDpath in glob.glob(os.path.join(self.RProot, 'FLUIDS', '*.FLD')):
            FLD = os.path.split(FLDpath)[1].split('.')[0]
            hash_ = None
            for line in open(FLDpath).readlines():
                if 'Hash' in line:
                    hash_ = line.split('!')[0].strip()
                    break
            for line in open(FLDpath).readlines():
                if 'CAS' in line:
                    o[hash_] = line.split('!')[0].strip()
                    break
        return o

    def get_RPlookup(self):
        o = {}
        for FLDpath in glob.glob(os.path.join(self.RProot, 'FLUIDS', '*.FLD')):
            FLD = os.path.split(FLDpath)[1].split('.')[0]
            hash_ = None
            for line in open(FLDpath).readlines():
                if 'Hash' in line:
                    hash_ = line.split('!')[0].strip()
                    break
            o[hash_] = FLD
        return o

    def get_binary_interaction(self):
        BIP = []
        iBNC = lines_equals("BNC", self.lines)[0]
        iend = iBNC
        while iend < iBNC + 10000:
            if self.lines[iend].strip() == '':
                break
            else:
                iend += 1
        for chunk in '\n'.join(self.lines[iBNC+1:iend]).split('!\n'):
            if all([line.strip().startswith('?') or not line.strip() for line in chunk.split('\n')]):
                continue
            else:
                chunklines = chunk.split('\n')
                Name1, Name2 = chunklines[0][1::].split('[')[0].strip().split('/')
                ilastcomment = lines_starts_with('?', chunklines)[-1]
                refrow = chunklines[1:ilastcomment+1]
                hash1, hash2 = [el.strip() for el in chunklines[ilastcomment+1].split('/')]
                def to_float(s):
                    try:
                        return float(s)
                    except:
                        return s
                # start at the bottom and work backwards to find the last mixture model definition
                for irow in reversed(range(len(chunklines)-1)):
                    model, betaT, gammaT, betaV, gammaV, Fij = [to_float(el) for el in chunklines[irow].split(' ') if el][0:6]
                    if model in (['PR1','ST1','TRN'] + [f'TC{i}' for i in range(1,8)] + [f'VC{i}' for i in range(1,8)]):
                        # print('XX', chunklines[irow])
                        pass
                    else:
                        # print('  ', chunklines[irow])
                        break

                try:
                    BIP.append(dict(
                        betaT=betaT,
                        gammaT=gammaT,
                        betaV=betaV,
                        gammaV=gammaV,
                        F=Fij,
                        BibTeX='?',
                        Name1=self.RPlookup[hash1],
                        Name2=self.RPlookup[hash2],
                        CAS1=self.CASlookup[hash1],
                        CAS2=self.CASlookup[hash2],
                        function=model
                    ))
                except KeyError as ke:
                    print(ke)
        return BIP

    def get_departure_functions(self):
        iMXM = lines_starts_with("#MXM", self.lines)
        departure_functions = []

        # This is a placeholder for XR0 models
        departure_functions.append(dict(
            type='GERG-2008',
            n = [],
            t = [],
            d = [],
            l = [],
            eta = [],
            beta = [],
            gamma = [],
            epsilon = [],
            Npower=0,
            Name='XR0',
            aliases=[],
            BibTeX='??'
        ))

        for istart in iMXM:
            iend = istart
            while iend < istart + 30:
                if self.lines[iend].strip() == '':
                    break
                else:
                    iend += 1
            blocklines = self.lines[istart:iend]
            code = blocklines[1].split(' ')[0]
            if code in ['XR0','LIN','TR1']:
                continue
            istart = lines_contains('Descriptors for ', blocklines)[0] + 2 # index of row with spec
            specnumbers = [int(el) for el in blocklines[istart].split("!")[0].split(' ') if el]
            Npoly, Mpoly, dummy, Nspecial, Mspecial, NGaussian, MGaussian = specnumbers[0:7]
            def split_line(line):
                return [float(el.replace('d','e')) for el in line.split('!')[0].split()]

            if Nspecial > 0:
                # These GERG terms are combinations of polynomial and special terms
                Nlines = Npoly + Nspecial
                if Mpoly == 4:
                    blockpoly = blocklines[istart+1:istart+1+Npoly]
                    npoly, tpoly, dpoly, lpoly = zip(*[split_line(line) for line in blockpoly])
                    assert(len(npoly)==Npoly)
                else:
                    if Mpoly != 0:
                        raise ValueError(Mpoly)

                if Mspecial == 7:
                    blockspe = blocklines[istart+1+Npoly: istart+1+Npoly+Nspecial]
                    nspe, tspe, dspe, eta, epsilon, beta, gamma  = zip(*[split_line(line) for line in blockspe])
                else:
                    nspe = [0.0]*Nspecial
                    tspe = [0.0]*Nspecial
                    dspe = [0.0]*Nspecial
                    eta = [0.0]*Nspecial
                    beta = [0.0]*Nspecial
                    gamma = [0.0]*Nspecial
                    epsilon = [0.0]*Nspecial
                    if Mspecial != 0:
                        raise ValueError(Mspecial)

                departure_functions.append(dict(
                    type='GERG-2008',
                    n = list(npoly) + list(nspe),
                    t = list(tpoly) + list(tspe),
                    d = list(dpoly) + list(dspe),
                    l = list(lpoly) + [0.0]*Nspecial,
                    eta = [0.0]*Npoly + list(eta),
                    beta = [0.0]*Npoly + list(beta),
                    gamma = [0.0]*Npoly + list(gamma),
                    epsilon = [0.0]*Npoly + list(epsilon),
                    Npower=Npoly,
                    Name=code,
                    aliases=[],
                    BibTeX='??'
                ))

            else:
                # Polynomial and Gaussian terms
                Nlines = Npoly + NGaussian
                blockpoly = blocklines[istart+1 : istart+1+Npoly]
                if Mpoly == 4:
                    npoly, tpoly, dpoly, lpoly = zip(*[split_line(line) for line in blockpoly])
                elif Mpoly == 3:
                    npoly, tpoly, dpoly = zip(*[split_line(line) for line in blockpoly])
                    lpoly = [0.0]*len(dpoly)
                else:
                    if Mpoly != 0:
                        raise ValueError(Mpoly)

                if MGaussian == 12:
                    blockgau = blocklines[istart+1+Npoly:istart+1+Npoly+NGaussian+1]
                    ngau, tgau, dgau, dum1gau, dum2gau, eta, beta, gamma, epsilon, dum3, dum4, dum5  = zip(*[split_line(line) for line in blockgau])
                else:
                    ngau = [0.0]*NGaussian
                    tgau = [0.0]*NGaussian
                    dgau = [0.0]*NGaussian
                    eta = [0.0]*NGaussian
                    beta = [0.0]*NGaussian
                    gamma = [0.0]*NGaussian
                    epsilon = [0.0]*NGaussian
                    if MGaussian != 0:
                        raise ValueError(MGaussian)

                departure_functions.append(dict(
                    type='Gaussian+Exponential',
                    n=list(npoly) + list(ngau),
                    t=list(tpoly) + list(tgau),
                    d=list(dpoly) + list(dgau),
                    l=list(lpoly) + [0.0]*NGaussian,
                    eta=[0.0]*Npoly + (-np.array(eta)).tolist(),
                    beta=[0.0]*Npoly + (-np.array(beta)).tolist(),
                    gamma=[0.0]*Npoly + list(gamma),
                    epsilon=[0.0]*Npoly + list(epsilon),
                    Npower=Npoly,
                    Name=code,
                    aliases=[],
                    BibTeX='??'
                ))

        return departure_functions

    def write(self, *, base):
        with open(base + '/mixture_departure_functions.json','w') as fp:
            fp.write(json.dumps(self.dep, indent=2))
        with open(base + '/mixture_binary_pairs.json','w') as fp:
            fp.write(json.dumps(self.BIP, indent=2))

HMX_header = """HMX               !Mnemonic for mixture model, must match hfmix on call to SETUP.
4                 !Version number

! Changelog:
! ---------

#BNC              !Binary mixing coefficients
BNC
? Binary mixing coefficients for the various mixing rules used with the HMX model:
?
? KWi:  (i = 1,2,3,...,A,B,...)  --->  Kunz-Wagner mixing rules
?   model     BetaT     GammaT    BetaV     GammaV    Fij      not used
?
!"""

HMX_footer = """
@END
c        1         2         3         4         5         6         7         8
c2345678901234567890123456789012345678901234567890123456789012345678901234567890
"""

BIN_template = """
?{name1}/{name2}                        [{name1}/{name2}]
?FITTED {annotation}
  {hash1}/{hash2}
    {model}     {betaT:14.11f} {gammaT:14.11f} {betaV:14.11f} {gammaV:14.11f} {Fij:14.11f}  0.             0. 0. 0. 0. 0. 0.
    TC5    0.0      0.0     0.0       0.             0.             0.             0. 0. 0. 0. 0. 0.
    VC5    0.0      0.0     0.0       0.             0.             0.             0. 0. 0. 0. 0. 0.
!"""

MODEL_template = """#MXM              !Mixture model specification
{model} {annotation}
?
!```````````````````````````````````````````````````````````````````````````````
 BetaT    GammaT   BetaV    GammaV    Fij    not used      !Descriptors for binary-specific parameters
  1.0      1.0      1.0      1.0      0.0      0.0         !Default values (i.e. ideal-solution)
  {Nexp} {Ntermsexp}      0        {NKW} {NtermsKW}      {NGaussian} {NtermsGaussian}      0 0      0 0         !# terms and # coefs/term for normal terms, Kunz-Wagner terms, and Gaussian terms.  3rd column is not used.
  """

class HMXBuilder:
    """
    Read in CoolProp-format departure term JSON structure and build a REFPROP-format HMX.BNC
    """
    def __init__(self, jbin, jdep):

        """ Determine acceptable names for functions """
        def get_function_names():
            function_names = set([el['Name'] for el in jdep])
            return function_names
        function_names = get_function_names()
        if len(function_names) > 100:
            raise ValueError("Max of 100 functions allowed")
        RP_function_names = {f:f'B{i:2d}'.replace(' ','0') for i,f in enumerate(function_names)}
        def get_hash(CAS, name):
            try:
                # Use CoolProp to obtain the standard inchi key
                AS = CP.AbstractState('HEOS', CAS)
                StdInChIkey = CP.get_fluid_param_string(CAS, 'INCHIKEY')
                uid = hashlib.sha256(StdInChIkey.encode('UTF-8')).hexdigest()[2:9] + '0'
            except ValueError:
                # Use REFPROP to do so
                import ctREFPROP.ctREFPROP as c
                RP = c.REFPROPFunctionLibrary(os.getenv('RPPREFIX'))
                r = RP.REFPROPdll(name,'','HASH',RP.MOLAR_BASE_SI,0,0,0,0,[1.0])
                assert(r.ierr == 0)
                uid = r.hUnits
            return uid

        out = HMX_header

        # The top part with the definitions of what model to use for each binary pair and the 
        # model name
        for el in jbin:
            if 'function' not in el:
                RP_function_number = 'BA0'
            else:
                if el['function'] == 'XR0':
                    RP_function_number = 'XR0'
                else:
                    RP_function_number = RP_function_names[el['function']]
            if 'xi' in el and 'zeta' in el:
                # convert values to gammaT and gammaV
                # yadda...
                print("TODO: Need to convert values to gammaT and gammaV......................................")
                # values = 1.0, el['gammaT'], 1.0, el['gammaV'], el['F']
                continue
            else:
                template_values = {
                    'betaT': el['betaT'], 
                    'gammaT': el['gammaT'], 
                    'betaV': el['betaV'], 
                    'gammaV': el['gammaV'], 
                    'Fij': el['F'],
                    'name1': el['Name1'],
                    'name2': el['Name2'],
                    'annotation': 'time/date',
                    'hash1': get_hash(el['CAS1'], el['Name1']),
                    'hash2': get_hash(el['CAS2'], el['Name2']),
                    'model': RP_function_number
                }
                entry = BIN_template.format(**template_values)
                # print(entry)
                out += entry

        out += '\n\n'

        for el in jdep:
            # Build the departure term
            # print(el)
            model = RP_function_names[el["Name"]]
            annotation = f'{el["Name"]} '

            Nexp = Ntermsexp = NKW = NtermsKW = NGaussian = NtermsGaussian = 0
            if el['type'] == 'Exponential':
                Nexp = len(el['n'])
                Ntermsexp = 4 if 'l' in el and isinstance(el['l'], list) else 3
                
                # print(Nexp, Ntermsexp)
                if Ntermsexp == 4:
                    n, t, d, l = el['n'], el['t'], el['d'], el['l']
                    rows = []
                    for i in range(len(t)):
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.10f} {l[i]:0.10f}')
                        if i == 0:
                            rows[-1] += f' ! n(i),t(i),d(i),l(i) in term n_i*tau^t_i*delta^d_i*exp(-delta^l_i)'
                elif Ntermsexp == 3:
                    n, t, d = el['n'], el['t'], el['d']
                    first_row = f'{n[0]} {t[0]} {d[0]:0.1f} ! n(i),t(i),d(i) in term n_i*tau^t_i*delta^d_i'
                    rows = []
                    for i in range(len(t)):
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.10f}')
                        if i == 0:
                            rows[-1] += f' ! n(i),t(i),d(i) in term n_i*tau^t_i*delta^d_i'
                else:
                    raise ValueError()

            elif el['type'] == 'GERG-2008':
                # print(el)
                Nexp = el['Npower']
                Ntermsexp = 3
                assert('l' not in el)
                NKW = len(el['n'])-Nexp
                NtermsKW = 7
                
                n, t, d, eta, epsilon, beta, gamma = el['n'], el['t'], el['d'], el['eta'], el['epsilon'], el['beta'], el['gamma']
                l = None
                rows = []
                if Nexp > 0:
                    for i in range(Nexp):
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.10f} ')
                        if i == 0:
                            rows[-1] += '! n(i),t(i),d(i) in term n_i*tau^t_i*delta^d_i'
                if NKW > 0:
                    for i in range(Nexp,len(t)):
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.10f} {eta[i]} {epsilon[i]} {beta[i]} {gamma[i]} ')
                        if i == Nexp:
                            rows[-1] += '! n(i),t(i),d(i),eta(i),epsilon(i),beta(i),gamma(i) in term n_i*tau^t_i*delta^d_i*exp(-eta*(delta-epsilon)^2-beta*(delta-gamma))'

            elif el['type'] == 'Gaussian+Exponential':
                # print(el)
                Nexp = el['Npower']
                NGaussian = len(el['n'])-Nexp
                
                n, t, d, l, eta, epsilon, beta, gamma = el['n'], el['t'], el['d'], el['l'], el['eta'], el['epsilon'], el['beta'], el['gamma']
                rows = []
                if Nexp > 0:
                    Ntermsexp = 4
                    for i in range(Nexp):
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.16f} {l[i]:0.16f} ')
                        if i == 0:
                            rows[-1] += '! n(i),t(i),d(i),l(i) in term n_i*tau^t_i*delta^d_i*exp(-sgn(l_i)*delta^l_i)'
                if NGaussian > 0:
                    NtermsGaussian = 12
                    for i in range(Nexp,len(t)):
                        negetai = -eta[i]
                        negbetai = -beta[i]
                        rows.append(f'{n[i]} {t[i]} {d[i]:0.16f} 2.0 2.0 {negetai} {negbetai} {gamma[i]} {epsilon[i]} 0.0 0.0 0.0 0.0')
                        if i == Nexp:
                            rows[-1] += '! n(i),t(i),d(i),_,_,eta(i),beta(i),gamma(i),epsilon(i),_,_,_,_ in term n_i*tau^t_i*delta^d_i*exp(eta*(delta-epsilon)^2+beta*(tau-gamma)^2)'
            else:
                raise KeyError(el['type'])
                
            out += MODEL_template.format(**locals())
            out += '\n'.join(rows) + '\n\n'

        out += '\n\n' + HMX_footer

        self.out = out

    def get(self):
        return self.out

def to_teqp(RProot, outroot):
    os.makedirs(os.path.join(outroot,'dev','mixtures'), exist_ok=True)
    os.makedirs(os.path.join(outroot,'dev','fluids'), exist_ok=True)
    hmx = HMXDeconstructor(RProot+'/FLUIDS/HMX.BNC', RProot)
    hmx.write(base=outroot+'/dev/mixtures')
    for fld in glob.glob(RProot+'/FLUIDS/*.FLD'):
        name = os.path.split(fld)[1].split('.')[0]
        try:
            FLD = FLDDeconstructor(fld).write_JSON(outroot+f'/dev/fluids/{name}.json')
        except BaseException as BE:
            print(fld, BE)
            # raise
            pass

def test_CoolProp(root):
    CP.set_config_bool(CP.OVERWRITE_FLUIDS, True)
    for path in glob.glob(os.path.join(root,'dev','fluids','*.json')):
        FLD = os.path.split(path)[1].split('.')[0]
        CP.add_fluids_as_JSON('HEOS', open(path).read())
        AS = CP.AbstractState('HEOS', FLD)
    CP.set_departure_functions(open(root+'/dev/mixtures/mixture_departure_functions.json').read())
    CP.set_interaction_parameters(open(root+'/dev/mixtures/mixture_binary_pairs.json').read())

def get_all_ancillaries():
    ancillaries = {}
    for fld in glob.glob('C:/Program Files (x86)/REFPROP/FLUIDS/*.FLD'):
        name = os.path.split(fld)[1].split('.')[0]
        FLD = FLDDeconstructor(fld)
        print(fld)
        FLD.get_EOS()
        anc = FLD.get_all_ancillaries()
        T = anc['DV']['T_r']*0.8
        FLD.check_ancillaries(anc, name, T)
        ancillaries[name] = anc

    with open('all_ancillaries.json','w') as fp:
        fp.write(json.dumps(ancillaries, indent=2))

if __name__ == '__main__':
    RProot = os.getenv('RPPREFIX')
    to_teqp(RProot, 'teqp')
    test_CoolProp('teqp')

    import ctREFPROP.ctREFPROP as c, os
    RP = c.REFPROPFunctionLibrary(os.environ['RPPREFIX'])
    print(RP.RPVersion())


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from conan import ConanFile
from conan.tools.cmake import CMake, cmake_layout
from conan.tools.build import can_run
import os


class TestPackageConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "CMakeToolchain", "CMakeDeps"

    def requirements(self):
        self.requires(self.tested_reference_str)

    def layout(self):
        cmake_layout(self)

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def test(self):
        if can_run(self):
            cmd = os.path.join(self.cpp.build.bindir, "test_package")
            self.run(cmd, env="conanrun")

        # If we are on conan 2 we can check the license info is populated
        if hasattr(self, 'dependencies'):
            catch2 = self.dependencies["catch2"]
            assert os.path.exists(f'{catch2.package_folder}/licenses/LICENSE.txt')
            assert catch2.license == 'BSL-1.0'



from . import autodiff4py

# List of all autodiff C++ number types exported to Python. Extend this as other
# number types are exported!
_autodiff_number_types = [
    autodiff4py.real1st,
    autodiff4py.real2nd,
    autodiff4py.real3rd,
    autodiff4py.real4th,
]

# -------------------------------------------------------------------------------------
# Define the __format__ methods for all autodiff number types. This is needed so
# that we can write formatted strings such as f"{x:.3f}" and avoid a Python
# runtime error.
def _autodiff_number_format(self, spec):
    return format(self.val(), spec)

for numbertype in _autodiff_number_types:
    numbertype.__format__ = _autodiff_number_format

# -------------------------------------------------------------------------------------
# The code below is needed so that autodiff numbers can be used in plotly figures.
# It teaches PlotlyJSONEncoder on how to encode an autodiff number to JSON.
try:
    from plotly.utils import PlotlyJSONEncoder
    plotly_default = PlotlyJSONEncoder.default

    def is_autodiff_number_type(o):
        return any (t for t in _autodiff_number_types if isinstance(o, t))

    def plotly_default_new(self, o):
        if is_autodiff_number_type(o):
            return o.val()
        else:
            return plotly_default(self, o)

    PlotlyJSONEncoder.default = plotly_default_new
except:
    pass
# -------------------------------------------------------------------------------------

from .autodiff4py import *
# from ._extensions.some_module import *



""" This script runs *inside* the docker container """

import subprocess, os, sys, shutil, glob, timeit

tic = timeit.default_timer()

# Wipe contents of output folder 
# This script is running inside the docker container, so this should be allowed
for g in glob.glob('/output/*'):
    if os.path.isdir(g):
        shutil.rmtree(g)
    else:
        os.remove(g)

EXE = '/teqp/build/catch_tests'

# Collect the list of tags to be run
all_tags = []
output = subprocess.run(f'{EXE} --list-tags', shell = True, stdout = subprocess.PIPE).stdout.decode('utf-8')
for il, line in enumerate(output.split('\n')[1::]):
    if not line or '[' not in line: continue
    tag = '[' + line.split('[')[1]
    all_tags.append(tag)

tag_times = {}
for tag in all_tags:
    root =  tag.replace('[', '').replace(']','') + '.txt'
    print(tag, ' --> ', root)

    cmd = f'timeout 60m valgrind --tool=memcheck --error-limit=no --track-origins=yes {EXE} ' + tag
    tic1 = timeit.default_timer()
    with open('/output/log_'+root,'w') as fp_stderr:
        with open('/output/err_'+root,'w') as fp_stdout:
            subprocess.run(cmd, shell = True, stdout = fp_stdout, stderr = fp_stderr)    
    toc1 = timeit.default_timer()
    tag_times[tag] = toc1-tic1

    # Copy a debugging file to the output if present
    src_memorytest = 'memorytest.txt'
    dest_memorytest = '/output/memorytest_'+root
    if os.path.exists(src_memorytest):
        shutil.move(src_memorytest, dest_memorytest)

    print(open('/output/log_'+root).readlines()[-1])

# Print times taken for each tag
print('Time (in sec) taken for each tag')
for k, v in tag_times.items():
    print(k, v)

# Store all the outputs in zip archive
os.makedirs('/output/errors')
os.makedirs('/output/ok')
for g in glob.glob('/output/log_*.txt'):
    test_name = os.path.split(g)[1].replace('log_','').replace('.txt','')
    errname = f'/output/err_{test_name:s}.txt'
    is_ok = '0 errors from 0 contexts' in open(g).read()
    dest = '/output/ok' if is_ok else '/output/errors'
    for name in [g, errname]:
        shutil.move(name, dest)
shutil.make_archive('/output/errors','zip','/output/errors')
shutil.make_archive('/output/ok','zip','/output/ok')

toc = timeit.default_timer()
print('elapsed time [s]:', toc-tic)


#!/usr/bin/env python3

#              Copyright Catch2 Authors
# Distributed under the Boost Software License, Version 1.0.
#   (See accompanying file LICENSE_1_0.txt or copy at
#        https://www.boost.org/LICENSE_1_0.txt)

# SPDX-License-Identifier: BSL-1.0

"""
This test script verifies that the random ordering of tests inside
Catch2 is invariant in regards to subsetting. This is done by running
the binary 3 times, once with all tests selected, and twice with smaller
subsets of tests selected, and verifying that the selected tests are in
the same relative order.
"""

import subprocess
import sys
import random
import xml.etree.ElementTree as ET

def none_to_empty_str(e):
    if e is None:
        return ""
    assert type(e) is str
    return e

def list_tests(self_test_exe, tags, rng_seed):
    cmd = [self_test_exe, '--reporter', 'xml', '--list-tests', '--order', 'rand',
            '--rng-seed', str(rng_seed)]
    tags_arg = ','.join('[{}]~[.]'.format(t) for t in tags)
    if tags_arg:
        cmd.append(tags_arg)
    process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if stderr:
        raise RuntimeError("Unexpected error output:\n" + process.stderr)

    root = ET.fromstring(stdout)
    result = [(none_to_empty_str(tc.find('Name').text),
               none_to_empty_str(tc.find('Tags').text),
               none_to_empty_str(tc.find('ClassName').text)) for tc in root.findall('./TestCase')]

    if len(result) < 2:
        raise RuntimeError("Unexpectedly few tests listed (got {})".format(
            len(result)))
    return result

def check_is_sublist_of(shorter, longer):
    assert len(shorter) < len(longer)
    assert len(set(longer)) == len(longer)

    indexes_in_longer = {s: i for i, s in enumerate(longer)}
    for s1, s2 in zip(shorter, shorter[1:]):
        assert indexes_in_longer[s1] < indexes_in_longer[s2], (
                '{} comes before {} in longer list.\n'
                'Longer: {}\nShorter: {}'.format(s2, s1, longer, shorter))

def main():
    self_test_exe, = sys.argv[1:]

    # We want a random seed for the test, but want to avoid 0,
    # because it has special meaning
    seed = random.randint(1, 2 ** 32 - 1)

    list_one_tag = list_tests(self_test_exe, ['generators'], seed)
    list_two_tags = list_tests(self_test_exe, ['generators', 'matchers'], seed)
    list_all = list_tests(self_test_exe, [], seed)

    # First, verify that restricting to a subset yields the same order
    check_is_sublist_of(list_two_tags, list_all)
    check_is_sublist_of(list_one_tag, list_two_tags)

if __name__ == '__main__':
    sys.exit(main())



from conans import ConanFile

class ArgsConan(ConanFile):
    name = "args"
    version = "6.3.0"
    url = "https://github.com/Taywee/args"
    description = "A simple header-only C++ argument parser library."
    license = "MIT"
    exports = ["LICENSE"]
    exports_sources = "args.hxx"

    def package(self):
        self.copy(pattern="LICENSE", dst="license")
        self.copy(pattern="args.hxx", dst="include")



#!/usr/bin/env python3

#              Copyright Catch2 Authors
# Distributed under the Boost Software License, Version 1.0.
#   (See accompanying file LICENSE_1_0.txt or copy at
#        https://www.boost.org/LICENSE_1_0.txt)

# SPDX-License-Identifier: BSL-1.0

from typing import List, Tuple

import os
import subprocess

def configure_and_build(source_path: str, project_path: str, options: List[Tuple[str, str]]):
    base_configure_cmd = ['cmake',
                          '-B{}'.format(project_path),
                          '-H{}'.format(source_path),
                          '-DCMAKE_BUILD_TYPE=Debug',
                          '-DCATCH_DEVELOPMENT_BUILD=ON']
    for option, value in options:
        base_configure_cmd.append('-D{}={}'.format(option, value))
    try:
        subprocess.run(base_configure_cmd,
                       stdout = subprocess.PIPE,
                       stderr = subprocess.STDOUT,
                       check = True)
    except subprocess.SubprocessError as ex:
        print("Could not configure build to '{}' from '{}'".format(project_path, source_path))
        print("Return code: {}".format(ex.returncode))
        print("output: {}".format(ex.output))
        raise
    print('Configuring {} finished'.format(project_path))

    build_cmd = ['cmake',
                 '--build', '{}'.format(project_path),
                 # For now we assume that we only need Debug config
                 '--config', 'Debug']
    try:
        subprocess.run(build_cmd,
                       stdout = subprocess.PIPE,
                       stderr = subprocess.STDOUT,
                       check = True)
    except subprocess.SubprocessError as ex:
        print("Could not build project in '{}'".format(project_path))
        print("Return code: {}".format(ex.returncode))
        print("output: {}".format(ex.output))
        raise
    print('Building {} finished'.format(project_path))

def run_and_return_output(base_path: str, binary_name: str, other_options: List[str]) -> Tuple[str, str]:
    # For now we assume that Windows builds are done using MSBuild under
    # Debug configuration. This means that we need to add "Debug" folder
    # to the path when constructing it. On Linux, we don't add anything.
    config_path = "Debug" if os.name == 'nt' else ""
    full_path = os.path.join(base_path, config_path, binary_name)

    base_cmd = [full_path]
    base_cmd.extend(other_options)

    try:
        ret = subprocess.run(base_cmd,
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             check = True,
                             universal_newlines = True)
    except subprocess.SubprocessError as ex:
        print('Could not run "{}"'.format(base_cmd))
        print('Args: "{}"'.format(other_options))
        print('Return code: {}'.format(ex.returncode))
        print('stdout: {}'.format(ex.stdout))
        print('stderr: {}'.format(ex.stdout))
        raise

    return (ret.stdout, ret.stderr)



import sys, ctypes as ct, json, timeit

def trim(s):
    return s.raw.replace(b'\x00',b'').strip().decode('utf-8')

class DLLCaller():
    def __init__(self, full_path):
        if sys.platform.startswith('win'):
            loader_fcn = ct.WinDLL
        else:
            loader_fcn = ct.CDLL

        self.dll = loader_fcn(full_path)

    def _getfcn(self, DLL, fname):
        try:
            return getattr(DLL, fname)
        except BaseException as BE:
            return None

    def build_model(self, model):
        f = self._getfcn(self.dll, 'build_model')
        hrf = ct.create_string_buffer(json.dumps(model).encode('utf-8'))
        uid = ct.c_longlong()
        errmsg = ct.create_string_buffer(1000)
        errcode = f(hrf, ct.byref(uid), errmsg, len(errmsg))
        if errcode == 0:
            return uid
        else:
            raise ValueError(trim(errmsg))

    def get_Arxy(self, *, uid, NT, ND, T, rho, z):
        f = self._getfcn(self.dll, 'get_Arxy')
        NT = ct.c_int(NT)
        ND = ct.c_int(ND)
        T = ct.c_double(T)
        rho = ct.c_double(rho)
        molefrac = (len(z)*ct.c_double)(*z)
        Ncomp = len(z)
        o = ct.c_double()
        errmsg = ct.create_string_buffer(1000)
        tic = timeit.default_timer()
        errcode = f(uid, NT, ND, T, rho, molefrac, Ncomp, ct.byref(o), errmsg, len(errmsg))
        toc = timeit.default_timer()
        if errcode == 0:
            return o.value
        else:
            raise ValueError(trim(errmsg))

if __name__ == '__main__':
    # Now load the library
    c = DLLCaller(full_path = '../../bld/Release/libteqpc.dylib') # or .dll on windows
    model = {
      'kind': 'vdW1',
      'model': {'a': 1, 'b': 2}
    }
    uid = c.build_model(model)
    print(c.get_Arxy(uid=uid, NT=0,ND=1,T=300,rho=1,z=[1.0]))


#!/usr/bin/env python3

#              Copyright Catch2 Authors
# Distributed under the Boost Software License, Version 1.0.
#   (See accompanying file LICENSE_1_0.txt or copy at
#        https://www.boost.org/LICENSE_1_0.txt)

# SPDX-License-Identifier: BSL-1.0

"""
This test script verifies that the testCasePartial{Starting,Ended} reporter
events fire properly. This is done by calling a test binary compiled with
reporter that reports specifically testCase* events, and verifying the
outputs match what we expect.
"""

import subprocess
import sys

expected_section_output = '''\
TestCaseStarting: section
TestCaseStartingPartial: section#0
TestCasePartialEnded: section#0
TestCaseStartingPartial: section#1
TestCasePartialEnded: section#1
TestCaseStartingPartial: section#2
TestCasePartialEnded: section#2
TestCaseStartingPartial: section#3
TestCasePartialEnded: section#3
TestCaseEnded: section
'''

expected_generator_output = '''\
TestCaseStarting: generator
TestCaseStartingPartial: generator#0
TestCasePartialEnded: generator#0
TestCaseStartingPartial: generator#1
TestCasePartialEnded: generator#1
TestCaseStartingPartial: generator#2
TestCasePartialEnded: generator#2
TestCaseStartingPartial: generator#3
TestCasePartialEnded: generator#3
TestCaseEnded: generator
'''


from typing import List

def get_test_output(test_exe: str, sections: bool) -> List[str]:
    cmd = [test_exe, '--reporter', 'partial']
    if sections:
        cmd.append('section')
    else:
        cmd.append('generator')

    ret = subprocess.run(cmd,
                         stdout = subprocess.PIPE,
                         stderr = subprocess.PIPE,
                         timeout = 10,
                         check = True,
                         universal_newlines = True)

    return ret.stdout

def main():
    test_exe, = sys.argv[1:]
    actual_section_output = get_test_output(test_exe, sections = True)

    assert actual_section_output == expected_section_output, (
    'Sections\nActual:\n{}\nExpected:\n{}\n'.format(actual_section_output, expected_section_output))

    actual_generator_output = get_test_output(test_exe, sections = False)
    assert actual_generator_output == expected_generator_output, (
    'Generators\nActual:\n{}\nExpected:\n{}\n'.format(actual_generator_output, expected_generator_output))



if __name__ == '__main__':
    sys.exit(main())



import os
import glob
import subprocess
import json

import numpy as np
import scipy.optimize
import matplotlib.pyplot as plt

class AncillaryFitter:
    
    def __init__(self, *, FLD, prop, RP):
        """
        FLD: the FLD name for REFPROP
        RP: the REFPROP instance
        prop: What to fit, options are {'rhoL','rhoV','pS'}
        """
        self.FLD = FLD
        Tt, Tc, Tr, Tmin, rho_c, p_c, rhor = RP.REFPROPdll('','','TTRP;TC;TRED;TMIN;DC;PC;DRED', RP.MOLAR_BASE_SI, 0,0,0,0,[1.0]).Output[0:7]
        self.Ts = np.linspace(Tmin, 0.99*Tc)
        self.pS = np.array([RP.REFPROPdll('','TQ','P',RP.MOLAR_BASE_SI,0,0,T,0,[1.0]).Output[0] for T in self.Ts])
        self.rhoL = np.array([RP.REFPROPdll('','TQ','D',RP.MOLAR_BASE_SI,0,0,T,0,[1.0]).Output[0] for T in self.Ts])
        self.rhoV = np.array([RP.REFPROPdll('','TQ','D',RP.MOLAR_BASE_SI,0,0,T,1,[1.0]).Output[0] for T in self.Ts])
        T_r = Tc if Tr==Tc else Tr
        rho_r = rho_c if Tr==Tc else rhor
        self.theta = 1-self.Ts/T_r

        using_tau_r = False
        if prop in ['pS','pL','pV']:
            self.LHS = np.log(self.pS/p_c)*self.Ts/T_r
            using_tau_r = True
            reducing_value = p_c
            description = 'P=Pc*exp[sum(Ni*Theta^ti)*Tred/T]; Theta=1-T/Tred'
        elif prop in ['rhoL']:
            self.LHS = (self.rhoL/rho_r)-1
            reducing_value = rho_r
            description = 'D=Dred*[1+sum(Ni*Theta^ti)]; Theta=1-T/Tred'
        elif prop in ['rhoV']:
            self.LHS = np.log(self.rhoV/rho_r)
            reducing_value = rho_r
            description = 'D=Dred*exp[sum(Ni*Theta^ti)]; Theta=1-T/Tred'
        else:
            raise KeyError(prop)

        self.base_anc = {
            'T_r': T_r,
            'type': prop + ('noexp' if prop =='rhoL' else ''),
            'description': description,
            'using_tau_r': using_tau_r,
            'reducing_value': reducing_value,
            'Tmin': Tt,
            'Tmax': T_r
        }
        # print(self.LHS)

    def plot_values(self):
        plt.plot(self.theta, self.LHS)
        plt.gca().set(xlabel=r'$\theta=1-T/T_{\rm red}$', ylabel='LHS')
        plt.show()

    def evaluator(self, *, anc, T):
        """
        # Most are in this form:
        'PS5':  'P=Pc*EXP[SUM(Ni*Theta^ti)*Tc/T]',
        'DL1':  'D=Dc*[1+SUM(Ni*Theta^ti)]',
        'DV3':  'D=Dc*EXP[SUM(Ni*Theta^ti)]',
        """
        T_r = anc['T_r']
        theta = 1-T/T_r
        RHS = sum([_n*theta**_t for _n, _t in zip(anc['n'], anc['t'])])
        if anc['using_tau_r']:
            RHS *= T_r/T
        
        if anc['type'] == 'rhoLnoexp':
            return anc['reducing_value']*(1+RHS)
        elif anc['type'] in ['pV','pL','pS','rhoV','rhoL']:
            return np.exp(RHS)*anc['reducing_value']
        else:
            raise KeyError(anc['type'])

    def get_ancillary(self, *, n, t):
        anc = self.base_anc.copy()
        anc['n'] = np.array(n).tolist()
        anc['t'] = np.array(t).tolist()
        return anc
    
    def get_n(self, t):
        """
        This function takes the exponents t_i and does the 
        least squares solve for the coefficients n_i
        """
        A = np.zeros((len(self.theta), len(t)))
        for i in range(len(t)):
            A[:, i] = self.theta**t[i]
        n = np.linalg.lstsq(A, self.LHS, rcond=None)[0]
        return n, A

    def split_objective(self, params):
        n, A = self.get_n(t=params)
        predicted = np.dot(A, n)
        err = (predicted - self.LHS)/self.LHS
        val = np.sum(err**2)**0.5
        return val
    
    def optimize(self, bounds):
        return scipy.optimize.differential_evolution(self.split_objective, bounds=bounds)

def ancillary_injecter(*, RP, FLD, teqp, plot=True):
    ierr = RP.SETFLUIDSdll(FLD)
    if ierr > 0:
        print(RP.ERRMSGdll(ierr))
        quit()

    with open(teqp) as fp:
        j = json.load(fp)

    if 'rhoL' in j['ANCILLARIES']:
        print('Ancillaries already exist for FLD')
        return

    for propkey in ['pS', 'rhoL', 'rhoV']:
        anc = AncillaryFitter(FLD=FLD, prop=propkey, RP=RP)
        # anc.plot_values()
        opt = scipy.optimize.differential_evolution(anc.split_objective, bounds=[(0.1, 5) for i in range(5)])
        n, A = anc.get_n(t=opt.x)
        err = np.dot(A, n)-anc.LHS
        a = anc.get_ancillary(n=n, t=opt.x)
        y = np.array([anc.evaluator(anc=a, T=T_) for T_ in anc.Ts])
        if plot:
            plt.plot(anc.Ts, getattr(anc, propkey)/y-1)
            plt.title(f'{propkey} / {os.path.basename(FLD)}')
            plt.show()
        j['ANCILLARIES'][propkey] = a
        
    with open(teqp, 'w') as fp:
        fp.write(json.dumps(j,indent=2))


#!/usr/bin/env python3

#              Copyright Catch2 Authors
# Distributed under the Boost Software License, Version 1.0.
#   (See accompanying file LICENSE.txt or copy at
#        https://www.boost.org/LICENSE_1_0.txt)

# SPDX-License-Identifier: BSL-1.0

import os
import subprocess
import sys


def build_project(sources_dir, output_base_path, catch2_path):
    build_dir = os.path.join(output_base_path, 'ctest-registration-test')
    config_cmd = ['cmake',
                  '-B', build_dir,
                  '-S', sources_dir,
                  f'-DCATCH2_PATH={catch2_path}',
                  '-DCMAKE_BUILD_TYPE=Debug']

    build_cmd = ['cmake',
                 '--build', build_dir,
                 '--config', 'Debug']

    try:
        subprocess.run(config_cmd,
                       capture_output = True,
                       check = True,
                       text = True)
        subprocess.run(build_cmd,
                       capture_output = True,
                       check = True,
                       text = True)
    except subprocess.CalledProcessError as err:
        print('Error when building the test project')
        print(f'cmd: {err.cmd}')
        print(f'stderr: {err.stderr}')
        print(f'stdout: {err.stdout}')
        exit(3)

    return build_dir



def get_test_names(build_path):
    # For now we assume that Windows builds are done using MSBuild under
    # Debug configuration. This means that we need to add "Debug" folder
    # to the path when constructing it. On Linux, we don't add anything.
    config_path = "Debug" if os.name == 'nt' else ""
    full_path = os.path.join(build_path, config_path, 'tests')


    cmd = [full_path, '--reporter', 'xml', '--list-tests']
    result = subprocess.run(cmd,
                            capture_output = True,
                            check = True,
                            text = True)

    import xml.etree.ElementTree as ET
    root = ET.fromstring(result.stdout)
    return [tc.text for tc in root.findall('TestCase/Name')]


def list_ctest_tests(build_path):
    old_path = os.getcwd()
    os.chdir(build_path)

    cmd = ['ctest', '-C', 'debug', '--show-only=json-v1']
    result = subprocess.run(cmd,
                            capture_output = True,
                            check = True,
                            text = True)
    os.chdir(old_path)

    import json

    ctest_response = json.loads(result.stdout)
    tests = ctest_response['tests']
    test_names = []
    for test in tests:
        test_command = test['command']
        # First part of the command is the binary, second is the filter.
        # If there are less, registration has failed. If there are more,
        # registration has changed and the script needs updating.
        assert len(test_command) == 2
        test_names.append(test_command[1])
        test_name = test_command[1]

    return test_names

def escape_catch2_test_name(name):
    for char in ('\\', ',', '[', ']'):
        name = name.replace(char, f"\\{char}")
    return name

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(f'Usage: {sys.argv[0]} path-to-catch2-cml output-path')
        exit(2)
    catch2_path = sys.argv[1]
    output_base_path = sys.argv[2]
    sources_dir = os.path.dirname(os.path.abspath(sys.argv[0]))

    build_path = build_project(sources_dir, output_base_path, catch2_path)

    catch_test_names = [escape_catch2_test_name(name) for name in get_test_names(build_path)]
    ctest_test_names = list_ctest_tests(build_path)

    mismatched = 0
    for catch_test in catch_test_names:
        if catch_test not in ctest_test_names:
            print(f"Catch2 test '{catch_test}' not found in CTest")
            mismatched += 1
    for ctest_test in ctest_test_names:
        if ctest_test not in catch_test_names:
            print(f"CTest test '{ctest_test}' not found in Catch2")
            mismatched += 1

    if mismatched:
        print(f"Found {mismatched} mismatched tests catch test names and ctest test commands!")
        exit(1)



import sys, ctypes as ct, json, timeit, os

contents = '''
{
  "EOS": [
    {
      "BibTeX_CP0": "",
      "BibTeX_EOS": "Thol-THESIS-2015",
      "STATES": {
        "reducing": {
          "T": 1.32,
          "T_units": "LJ units",
          "rhomolar": 0.31,
          "rhomolar_units": "LJ units"
        }
      },
      "T_max": 1200,
      "T_max_units": "LJ units",
      "Ttriple": 290.25,
      "Ttriple_units": "LJ units",
      "alphar": [
        {
          "d": [4, 1, 1, 2, 2, 3, 1, 1, 3, 2, 2, 5],
          "l": [0, 0, 0, 0, 0, 0, 1, 2, 2, 1, 2, 1],
          "n": [0.52080730e-2, 0.21862520e+1, -0.21610160e+1, 0.14527000e+1, -0.20417920e+1, 0.18695286e+0, -0.62086250e+0, -0.56883900e+0, -0.80055922e+0, 0.10901431e+0, -0.49745610e+0, -0.90988445e-1],
          "t": [1.000, 0.320, 0.505, 0.672, 0.843, 0.898, 1.205, 1.786, 2.770, 1.786, 2.590, 1.294],
          "type": "ResidualHelmholtzPower"
        },
        {
          "beta": [0.625, 0.638, 3.91, 0.156, 0.157, 0.153, 1.16, 1.73, 383, 0.112, 0.119],
          "d": [1, 1, 2, 3, 3, 2, 1, 2, 3, 1, 1],
          "epsilon": [ 0.2053, 0.409, 0.6, 1.203, 1.829, 1.397, 1.39, 0.539, 0.934, 2.369, 2.43],
          "eta": [2.067, 1.522, 8.82, 1.722, 0.679, 1.883, 3.925, 2.461, 28.2, 0.753, 0.82],
          "gamma": [0.71, 0.86, 1.94, 1.48, 1.49, 1.945, 3.02, 1.11, 1.17, 1.33, 0.24],
          "n": [-0.14667177e+1, 0.18914690e+1, -0.13837010e+0, -0.38696450e+0, 0.12657020e+0, 0.60578100e+0, 0.11791890e+1, -0.47732679e+0, -0.99218575e-1, -0.57479320e+0, 0.37729230e-2],
          "t": [2.830, 2.548, 4.650, 1.385, 1.460, 1.351, 0.660, 1.496, 1.830, 1.616, 4.970],
          "type": "ResidualHelmholtzGaussian"
        }
      ],
      "gas_constant": 1.0,
      "gas_constant_units": "LJ units",
      "molar_mass": 1.0,
      "molar_mass_units": "LJ units",
      "p_max": 100000,
      "p_max_units": "LJ units",
      "pseudo_pure": false
    }
  ],
  "INFO":{
    "NAME": "LennardJones",
    "REFPROP_NAME": "LJF",
    "CAS": "N/A"
    }
}
'''

def trim(s):
    return s.raw.replace(b'\x00',b'').strip().decode('utf-8')

class DLLCaller():
    def __init__(self, full_path):
        if sys.platform.startswith('win'):
            loader_fcn = ct.WinDLL
        else:
            loader_fcn = ct.CDLL

        self.dll = loader_fcn(full_path)

    def _getfcn(self, DLL, fname):
        try:
            return getattr(DLL, fname)
        except BaseException as BE:
            return None

    def build_model(self, model):
        f = self._getfcn(self.dll, 'build_model')
        hrf = ct.create_string_buffer(json.dumps(model).encode('utf-8'))
        uid = ct.create_string_buffer(200)
        errmsg = ct.create_string_buffer(1000)
        errcode = f(hrf, uid, errmsg, len(errmsg))
        if errcode == 0:
            return uid
        else:
            raise ValueError(trim(errmsg))

    def get_Arxy(self, *, uid, NT, ND, T, rho, z):
        f = self._getfcn(self.dll, 'get_Arxy')
        NT = ct.c_int(NT)
        ND = ct.c_int(ND)
        T = ct.c_double(T)
        rho = ct.c_double(rho)
        molefrac = (len(z)*ct.c_double)(*z)
        Ncomp = len(z)
        o = ct.c_double()
        errmsg = ct.create_string_buffer(1000)
        tic = timeit.default_timer()
        errcode = f(uid, NT, ND, T, rho, molefrac, Ncomp, ct.byref(o), errmsg, len(errmsg))
        toc = timeit.default_timer()
        if errcode == 0:
            return o
        else:
            raise ValueError(trim(errmsg))

if __name__ == '__main__':
    # Now load the library
    c = DLLCaller(full_path = '../../bld/Release/teqpc.dll')
    model = {
      'kind': 'vdW1',
      'model': {'a': 1, 'b': 2}
    }
    uid = c.build_model(model)
    print(c.get_Arxy(uid=uid, NT=0,ND=1,T=300,rho=1,z=[1.0]))

    model = {
      'kind': 'multifluid',
      'model': {'components': [json.loads(contents)], 'departure': [], 'BIP': []}
    }
    uid = c.build_model(model)
    print(c.get_Arxy(uid=uid, NT=0, ND=0,T=1.5,rho=0.3,z=[1.0]))
    print(c.get_Arxy(uid=uid, NT=0, ND=1,T=1.5,rho=0.3,z=[1.0]))

    with open('ljf.json','w') as fp:
      fp.write(contents)

    model = {
      'kind': 'multifluid',
      'model': {'components': [json.load(open('ljf.json'))], 'departure': [], 'BIP': []}
    }
    uid = c.build_model(model)
    print(c.get_Arxy(uid=uid, NT=0, ND=0,T=1.5,rho=0.3,z=[1.0]))
    print(c.get_Arxy(uid=uid, NT=0, ND=1,T=1.5,rho=0.3,z=[1.0]))
    print(trim(uid))
    os.remove('ljf.json')


import os
import re
from conans import load, tools, ConanFile, CMake


def get_version():
    try:
        version = os.getenv('PROJECT_VERSION', None)
        if version:
            return version

        content = load('CMakeLists.txt')
        version = re.search('set\(PROJECT_VERSION (.*)\)', content).group(1)
        return version.strip()
    except:
        return None

class JsonSchemaValidatorConan(ConanFile):
    name = 'JsonSchemaValidator'
    version = get_version()
    url = 'https://github.com/pboettch/json-schema-validator'
    license = 'MIT'
    settings = 'os', 'compiler', 'build_type', 'arch'
    options = {
        'shared': [True, False],
        'fPIC': [True, False],
        'build_examples': [True, False],
        'build_tests': [True, False]
    }
    default_options = {
        'shared': False,
        'fPIC': True,
        'build_examples': True,
        'build_tests': False
    }
    generators = "CMakeDeps"
    exports_sources = [
        'CMakeLists.txt',
        'nlohmann_json_schema_validatorConfig.cmake.in',
        'src/*',
        'app/*',
        'test/*',
    ]
    requires = (
        'nlohmann_json/3.11.2'
    )
    _cmake = None

    def _configure_cmake(self):
        if self._cmake:
            return self._cmake
        self._cmake = CMake(self)
        self._cmake.definitions['JSON_VALIDATOR_BUILD_EXAMPLES'] = self.options.build_examples
        self._cmake.definitions['JSON_VALIDATOR_BUILD_TESTS'] = self.options.build_tests
        self._cmake.configure()
        return self._cmake

    def layout(self):
        build_type = str(self.settings.build_type).lower()
        self.folders.build = "build-{}".format(build_type)

    def build(self):
        cmake = self._configure_cmake()
        cmake.configure()
        cmake.build()

    def package(self):
        cmake = self._configure_cmake()
        cmake.install()

    def package_info(self):
        includedir = os.path.join(self.package_folder, "include")
        self.cpp_info.includedirs = [includedir]

        libdir = os.path.join(self.package_folder, "lib")
        self.cpp_info.libdirs = [libdir]
        self.cpp_info.libs += tools.collect_libs(self, libdir)

        bindir = os.path.join(self.package_folder, "bin")
        self.output.info("Appending PATH environment variable: {}".format(bindir))
        self.env_info.PATH.append(bindir)

        self.user_info.VERSION = self.version
